#
# MplHelp.py -- help classes for Matplotlib drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from ginga.misc import Bunch, Callback
from ginga.fonts import font_asst

import matplotlib.textpath as textpath


def load_font(font_name, font_file):
    from matplotlib import font_manager
    # may raise an exception
    font_manager.fontManager.addfont(font_file)


def get_cached_font(font_name, font_size):
    key = ('mpl', font_name)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        pass

    # font not loaded? try and load it
    try:
        info = font_asst.get_font_info(font_name, subst_ok=False)
        load_font(font_name, info.font_path)
        font_asst.add_cache(key, font_name)
        return font_name

    except Exception as e:
        # couldn't load font
        pass

    # try and substitute one of the built in fonts
    try:
        info = font_asst.get_font_info(font_name, subst_ok=True)
        load_font(font_name, info.font_path)

        font_asst.add_cache(key, font_name)
        return font_name

    except Exception as e:
        # couldn't load substitute font
        pass

    return font_name


class MplContext(object):

    def __init__(self, axes):
        self.axes = axes
        self.kwdargs = dict()
        self.stack = []

    def set_canvas(self, axes):
        self.axes = axes

    def init(self, **kwdargs):
        self.kwdargs = dict()
        self.kwdargs.update(kwdargs)

    def set(self, **kwdargs):
        self.kwdargs.update(kwdargs)

    def push(self, allow=[]):
        self.stack.append(self.kwdargs.copy())
        d = {name: self.kwdargs[name]
             for name in allow if name in self.kwdargs}
        self.kwdargs = d

    def pop(self):
        self.kwdargs = self.stack.pop()

    def update_fill(self, fill):
        if fill is None:
            self.kwdargs['fill'] = False
            return

        self.kwdargs['fill'] = True
        self.kwdargs['facecolor'] = fill._color_4tup

    def update_line(self, line):
        if line is not None:
            self.kwdargs['color'] = line._color_4tup
            self.kwdargs['linewidth'] = line.linewidth
            style = 'dashdot' if line.linestyle == 'dash' else 'solid'
            self.kwdargs['linestyle'] = style

    def update_patch(self, line, fill):
        self.update_fill(fill)

        if self.kwdargs['fill']:
            line_color_attr = 'facecolor'
            if 'facecolor' in self.kwdargs:
                line_color_attr = 'edgecolor'
        else:
            line_color_attr = 'color'

        if line is not None:
            self.kwdargs[line_color_attr] = line._color_4tup
            self.kwdargs['linewidth'] = line.linewidth
            style = 'dashdot' if line.linestyle == 'dash' else 'solid'
            self.kwdargs['linestyle'] = style

    def get_fontdict(self, font, fill):
        _font = get_cached_font(font.fontname, font.fontsize)
        fontdict = dict(color=fill._color_4tup, family=_font,
                        size=font.fontsize, transform=None)
        return fontdict

    def text_extents(self, text, font):
        # This is not completely accurate because it depends a lot
        # on the renderer used, but that is complicated under Mpl
        t = textpath.TextPath((0, 0), text, size=font.fontsize,
                              prop=font.fontname)
        bb = t.get_extents()
        wd, ht = bb.width, bb.height
        return (wd, ht)


class Timer(Callback.Callbacks):
    """Abstraction of a GUI-toolkit implemented timer."""

    def __init__(self, duration=0.0, mplcanvas=None):
        """Create a timer set to expire after `duration` sec.
        """
        super(Timer, self).__init__()

        self.duration = duration
        # For storing aritrary data with timers
        self.data = Bunch.Bunch()

        self._timer = mplcanvas.new_timer()
        self._timer.single_shot = True
        self._timer.add_callback(self._redirect_cb)

        for name in ('expired', 'canceled'):
            self.enable_callback(name)

    def start(self, duration=None):
        """Start the timer.  If `duration` is not None, it should
        specify the time to expiration in seconds.
        """
        if duration is None:
            duration = self.duration

        self.set(duration)

    def set(self, duration):

        self.stop()

        # Matplotlib timer set in milliseconds
        time_ms = int(duration * 1000.0)
        self._timer.interval = time_ms
        self._timer.start()

    def _redirect_cb(self):
        self.make_callback('expired')

    def stop(self):
        try:
            self._timer.stop()
        except Exception:
            pass

    def cancel(self):
        """Cancel this timer.  If the timer is not running, there
        is no error.
        """
        self.stop()
        self.make_callback('canceled')

    clear = cancel

# END



#
# AggHelp.py -- help classes for the Agg drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import aggdraw as agg

from ginga.fonts import font_asst


def get_cached_font(fontname, fontsize, color, alpha):
    key = ('agg', fontname, fontsize, color, alpha)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        # see if we can build the font
        info = font_asst.get_font_info(fontname, subst_ok=True)

        font = agg.Font(color, info.font_path, size=fontsize, opacity=alpha)
        font_asst.add_cache(key, font)

        return font


def load_font(font_name, font_file):
    # try to load it as a sanity check
    #agg.Font('black', font_file, size=10, opacity=255)

    if not font_asst.have_font(font_name):
        font_asst.add_font(font_file, font_name=font_name)
    return font_name


class AggContext(object):

    def __init__(self, canvas):
        self.canvas = canvas

    def set_canvas(self, canvas):
        self.canvas = canvas

    def _get_font(self, font, fill):
        color = fill.render.color[:3]
        op = fill.render.color[3]

        font = get_cached_font(font.fontname, font.fontsize, color, op)
        return font

    def text_extents(self, text, _font):
        # _font is an Agg font
        wd, ht = self.canvas.textsize(text, _font)
        return wd, ht

# END



# -*- coding: iso-8859-1 -*-
"""
Build a standalone application for Mac OS X and MS Windows platforms

 Usage (Mac OS X):
     python setup.py py2app

 Usage (Windows):
     python setup.py py2exe
"""
import sys
from setuptools import setup

info_plist_template = u"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>CFBundleName</key>
	<string>Ginga</string>
	<key>CFBundlePackageType</key>
	<string>APPL</string>
	<key>CFBundleGetInfoString</key>
	<string>Copyright © 2010-2016, Eric Jeschke (eric@naoj.org)</string>
	<key>CFBundleIconFile</key>
	<string>Ginga.icns</string>
	<!-- Version number - appears in About box -->
	<key>CFBundleShortVersionString</key>
	<string>%(version)s</string>
	<!-- Build number - appears in About box -->
	<key>CFBundleVersion</key>
	<string>%(build)s</string>
	<!-- Copyright notice - apears in About box -->
	<key>NSHumanReadableCopyright</key>
	<string>Copyright © 2010-2016, Eric Jeschke (eric@naoj.org)</string>
	<!-- Globally unique identifier -->
	<key>CFBundleIdentifier</key>
	<string>org.naoj.Ginga</string>
	<key>CFBundleDevelopmentRegion</key>
	<string>English</string>
	<key>CFBundleExecutable</key>
	<string>Ginga</string>
	<key>CFBundleDisplayName</key>
	<string>Ginga</string>
</dict>
</plist>
"""

from ginga import __version__

d = dict(version=__version__, build=__version__.replace('.', ''))
plist = info_plist_template % d

with open('Info.plist', 'w') as out_f:
    out_f.write(plist)


APP = ['Ginga.py']
DATA_FILES = []

OPTIONS = {'argv_emulation': True,
           'compressed': True,
           #'packages': 'ginga,scipy,numpy,kapteyn,astropy,PIL,matplotlib',
           'packages': 'ginga,scipy,numpy,astropy,PIL,matplotlib',
           'includes': ['sip', 'PyQt4._qt',],
           # currently creating some problems with the app build on mac os x
           # so exclude
           'excludes': ['cv2',],
           'matplotlib_backends': 'Qt4Agg',
           }

if sys.platform == 'darwin':
    # mac-specific options
    OPTIONS['plist'] = 'Info.plist'
    OPTIONS['iconfile'] = 'Ginga.icns'
    extra_options = dict(
        setup_requires=['py2app'],
        options={'py2app': OPTIONS},
    )

elif sys.platform == 'win32':
    extra_options = dict(
        setup_requires=['py2exe'],
        options={'py2exe': OPTIONS},
    )
else:
    extra_options = dict(
        # Normally unix-like platforms will use "setup.py install"
        # and install the main script as such
        scripts=["ginga"],
    )

    setup_requires=['py2app'],

setup(
    name="Ginga",
    app=APP,
    data_files=DATA_FILES,
    **extra_options
)



#
# draw.py -- mode for drawing on canvases
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""Draw Mode enables bindings that can facilitate drawing.

Enter the mode by
-----------------
* Space, then "b"

Exit the mode by
----------------
* Esc

Default bindings in mode
------------------------
* v : add a vertex to a polygon object, while drawing or editing a canvas
* z : delete a vertex from a polygon object, while drawing or editing a canvas
* x : delete a selected object while editing a canvas
* left drag : draw on the canvas in the defined shape
* right drag : draw on the canvas in the defined shape

"""
from ginga.modes.mode_base import Mode


class DrawMode(Mode):

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_draw=['__b', None, 'draw'],

            # NOTE: these only here to generate events. The events are
            # handled by the DrawingCanvas mixin
            kp_poly_add=['v', 'draw+v'],
            kp_poly_del=['z', 'draw+z'],
            kp_edit_del=['draw+x'],

            ms_draw=['draw+left', 'win+left', 'right'])

        ## bm = viewer.get_bindmap()
        ## bm.add_mode('__b', str(self), mode_type='locked', msg=None)

        ## bd = viewer.get_bindings()
        ## bd.merge_actions(self.viewer, bm, self, actions.items())

    def __str__(self):
        return 'draw'

    def start(self):
        self.viewer.switch_cursor('draw')

    def stop(self):
        self.viewer.switch_cursor('pick')
        self.onscreen_message(None)



#
# AutoCuts.py -- class for calculating auto cut levels
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga import trcalc
from ginga.misc import Bunch
#from ginga.misc.ParamSet import Param

from astropy.visualization import ZScaleInterval

have_scipy = True
autocut_methods = ['minmax', 'median', 'histogram', 'stddev', 'zscale']
try:
    import scipy.ndimage.filters
except ImportError:
    have_scipy = False
    autocut_methods = ['minmax', 'histogram', 'stddev', 'zscale']

__all__ = ['AutoCutsBase', 'Minmax', 'Histogram', 'StdDev', 'MedianFilter',
           'ZScale', 'get_autocuts', 'get_autocuts_names']


class Param(Bunch.Bunch):
    pass


class AutoCutsError(Exception):
    pass


class AutoCutsBase(object):
    """Base class for auto cuts algorithms."""

    @classmethod
    def get_params_metadata(cls):
        return []

    def __init__(self, logger):
        super().__init__()

        self.logger = logger
        self.kind = 'base'
        self.crop_radius = 512
        self.max_sample = 20000
        self.pct_sample = 0.02
        self.min_sample = 1000

    def update_params(self, **param_dict):
        # TODO: find a cleaner way to update these
        self.__dict__.update(param_dict)

    def get_algorithms(self):
        """Return the list of autocuts algorithms.

        Parameters
        ----------
        None

        Returns
        -------
        names: tuple of str
            The names of the autocut algorithms
        """
        return autocut_methods

    def get_crop(self, image, crop_radius=None):
        """Get a cropped region of data from an image.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.BaseImage`
            Image object from which the cut levels should be calculated

        crop_radius : int (optional, default `None`)
            The radius of a crop region to extract from the image.
            If `None`, then the radius will default to the `crop_radius`
            attribute.

        Returns
        -------
        result : ndarray
            The cropped data

        """
        # Even with numpy, it's kind of slow for some of the autocut
        # methods on a large image, so in those cases we can optionally
        # take a crop of size (radius*2)x(radius*2) from the center of
        # the image and calculate the cut levels on that
        if crop_radius is None:
            crop_radius = self.crop_radius

        wd, ht = image.get_size()
        (data, x1, y1, x2, y2) = image.cutout_radius(wd // 2, ht // 2,
                                                     crop_radius)
        return data

    def get_crop_data(self, data, crop_radius=None):
        """Get a cropped region of data from a ndarray.

        Parameters
        ----------
        data : ndarray
            Image data from which the data should be cropped

        crop_radius : int (optional, default `None`)
            The radius of a crop region to extract from the image data.
            If `None`, then the radius will default to the `crop_radius`
            attribute.

        Returns
        -------
        result : ndarray
            The cropped data

        """
        if crop_radius is None:
            crop_radius = self.crop_radius

        ht, wd = data.shape[:2]
        (data, x1, y1, x2, y2) = trcalc.cutout_radius(data, wd // 2, ht // 2,
                                                      crop_radius)
        return data

    def get_full(self, image, px_limit=None):
        """Get the full (or cropped) array of data from an image.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.BaseImage`
            Image object from which the cut levels should be calculated

        px_limit : int (optional, default `None`)
            The limit for extracting the full image data.  If the number
            of pixels in the data is larger than this value, a crop from
            the data will be used instead.

        Returns
        -------
        result : ndarray
            The (possibly cropped) data

        """
        wd, ht = image.get_size()
        num_px = wd * ht

        if px_limit is not None and num_px > px_limit:
            self.logger.info(f"size ({num_px}) > px limit ({px_limit}); "
                             "falling back to crop")
            return self.get_crop(image)

        data = image.cutout_data(0, 0, wd, ht)
        return data

    def get_sample(self, image, num_points=None):
        """Return a sample from the full data array of the passed image.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.BaseImage`
            Image object from which the cut levels should be calculated

        num_points : int (optional, default `None`)
            Specifies the number of points to sample.  If `None`, the number
            pixels will be calculated to a "reasonable representative sample".

        Returns
        -------
        result : ndarray
            The sampled data

        """
        wd, ht = image.get_size()
        total_points = wd * ht
        if num_points is None:
            num_points = min(max(self.min_sample,
                                 int(total_points * self.pct_sample)),
                             self.max_sample)
        num_points = min(num_points, total_points)
        if num_points == 0:
            return np.zeros((0, 0))

        # sample the data
        xmax = wd - 1
        ymax = ht - 1
        # evenly spaced sampling over rows and cols
        xskip = int(max(1.0, np.sqrt(xmax * ymax / float(num_points))))
        yskip = xskip

        cutout = image.cutout_data(0, 0, xmax, ymax,
                                   xstep=xskip, ystep=yskip)
        return cutout

    def get_sample_data(self, data, num_points=None):
        """Return a sample from a data array.

        Parameters
        ----------
        data : ndarray
            Image data from which the data should be sampled

        num_points : int (optional, default `None`)
            Specifies the number of points to sample.  If `None`, the number
            pixels will be calculated to a "reasonable representative sample".

        Returns
        -------
        result : ndarray
            The sampled data

        """
        ht, wd = data.shape[:2]
        total_points = wd * ht
        if num_points is None:
            num_points = min(max(self.min_sample,
                                 int(total_points * self.pct_sample)),
                             self.max_sample)
        num_points = min(num_points, total_points)
        if num_points == 0:
            return np.zeros((0, 0))

        # sample the data
        xmax = wd - 1
        ymax = ht - 1
        # evenly spaced sampling over rows and cols
        xskip = int(max(1.0, np.sqrt(xmax * ymax / float(num_points))))
        yskip = xskip

        cutout = trcalc.cutout_data(data, 0, 0, xmax, ymax,
                                    xstep=xskip, ystep=yskip)
        return cutout

    def calc_cut_levels(self, image):
        """Calculate the cut levels of an image according to the parameters.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.BaseImage`
            Image object from which the cut levels should be calculated

        Returns
        -------
        cut_levels : tuple of float (loval, hival)
            The cut levels that were calculated
        """
        self.logger.warning("Subclass should override this abstract method!")

    get_autocut_levels = calc_cut_levels

    def calc_cut_levels_data(self, data_np):
        """Calculate the cut levels of a ndarray according to the parameters.

        Parameters
        ----------
        data_np : ndarray
            A numpy array of data from which the cut levels should be
            calculated

        Returns
        -------
        cut_levels : tuple of float (loval, hival)
            The cut levels that were calculated
        """
        self.logger.warning("Subclass should override this abstract method!")

    def cut_levels(self, data_np, loval, hival, vmin=0.0, vmax=255.0):
        """Apply the cut levels to data.

        Parameters
        ----------
        data_np : ndarray
            A numpy array of data to which the cut levels should be applied

        loval : float
            The low cut level

        hival : float
            The high cut level

        vmin : float (optional, default 0.0)
            The floor of the range which is the output of the cut levels

        vmax : float (optional, default 255.0)
            The ceiling of the range which is the output of the cut levels

        Returns
        -------
        result : ndarray
            The result of applying the cut levels to the input array

        """
        loval, hival = float(loval), float(hival)
        # ensure hival >= loval
        hival = max(loval, hival)
        self.logger.debug("loval=%.2f hival=%.2f" % (loval, hival))
        delta = hival - loval
        if delta > 0.0:
            f = (((data_np - loval) / delta) * vmax)
            # NOTE: optimization using in-place outputs for speed
            f.clip(vmin, vmax, out=f)
            return f

        # hival == loval, so thresholding operation
        f = (data_np - loval).clip(vmin, vmax)
        f[f > 0.0] = vmax
        return f

    def __str__(self):
        return self.kind


class Clip(AutoCutsBase):
    """Calculate the cut levels as the minimum and maximum of the data.

    The calculation is:
        loval = min(sample_data)
        hival = max(sample_data)

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    """

    def __init__(self, logger):
        super().__init__(logger)
        self.kind = 'clip'

    def calc_cut_levels(self, image):
        loval, hival = image.get_minmax()

        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        loval = np.nanmin(data_np)
        hival = np.nanmax(data_np)

        return float(loval), float(hival)

    def cut_levels(self, data, loval, hival, vmin=0.0, vmax=255.0):
        return data.clip(vmin, vmax)


class Minmax(AutoCutsBase):
    """Calculate the cut levels as the minimum and maximum of the data.

    The calculation is:
        loval = min(sample_data)
        hival = max(sample_data)

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    """

    def __init__(self, logger):
        super().__init__(logger)
        self.kind = 'minmax'

    def calc_cut_levels(self, image):
        """See subclass documentation."""
        loval, hival = image.get_minmax()

        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        """See subclass documentation."""
        data = data_np[np.isfinite(data_np)]
        if data.size == 0:
            return (0, 0)
        loval = np.min(data)
        hival = np.max(data)

        return float(loval), float(hival)


class Histogram(AutoCutsBase):
    """Calculate the cut levels based on a histogram analysis of the data.

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    sample : str (optional, 'crop', 'grid', or 'full', default: 'crop')
        Specifies how to access the image for calculation:
        - crop: crop an area from the middle of the image
        - grid: sample data in a grid pattern across the image
        - full: use the full image data

    full_px_limit : int (optional, defaults to 1M)
        Specifies the limit for using the full data if sample == 'full'.
        If the number of pixels in the image is larger than this, then
        the image will fall back to using a crop.

    num_points : int (optional, defaults to None)
        Specifies the number of points in the grid if sample == 'grid',
        or the diameter of the crop, if sample == 'crop' (or 'full' and
        number of pixels exceeds `full_px_limit`).  If None, the number
        pixels will be calculated to a "reasonable representative sample".

    pct : float (optional, range: 0.0 - 1.0, defaults to 0.999)
        Specifies the percentage of the histogram bins to retain

    numbins : int (optional, defaults to 2048)
        Specifies the number of bins used in calculating the histogram

    """
    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='sample', type=str,
                  valid=['crop', 'grid', 'full'],
                  default='crop',
                  description="How to access data for calculation"),
            Param(name='full_px_limit', type=int,
                  default=1 * 1024 * 1024, allow_none=True,
                  description="For sample=full, fall back to crop if num_px > limit"),
            Param(name='num_points', type=int,
                  default=None, allow_none=True,
                  description="Number of points to sample (for sample=grid); 'None' for calculated default"),
            Param(name='pct', type=float,
                  widget='spinfloat', incr=0.001,
                  min=0.0, max=1.0, default=0.999,
                  description="Percentage of the histogram to retain"),
            Param(name='numbins', type=int,
                  min=100, max=10000, default=2048,
                  description="Number of bins for the histogram"),
        ]

    # NOTE: `usecrop` kwarg to be deprecated--accepted but not used
    # for backward compatibility with saved older settings
    def __init__(self, logger, usecrop=None, sample='crop',
                 full_px_limit=None, num_points=None,
                 pct=0.999, numbins=2048):
        super().__init__(logger)

        self.kind = 'histogram'
        self.sample = sample
        self.full_px_limit = full_px_limit
        self.num_points = num_points
        self.pct = pct
        self.numbins = numbins

    def calc_cut_levels(self, image):
        """See subclass documentation."""
        if self.sample == 'crop':
            if self.num_points is None:
                crop_radius = self.crop_radius
            else:
                crop_radius = int(self.num_points // 2)
            data = self.get_crop(image, crop_radius=crop_radius)
        elif self.sample == 'grid':
            data = self.get_sample(image, num_points=self.num_points)
        else:
            data = self.get_full(image, px_limit=self.full_px_limit)

        bnch = self.calc_histogram(data, pct=self.pct, numbins=self.numbins)
        loval, hival = bnch.loval, bnch.hival
        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        """See subclass documentation."""
        if self.sample == 'crop':
            if self.num_points is None:
                crop_radius = self.crop_radius
            else:
                crop_radius = int(self.num_points // 2)
            data = self.get_crop_data(data_np, crop_radius=crop_radius)
        elif self.sample == 'grid':
            data = self.get_sample_data(data_np, num_points=self.num_points)
        else:
            data = data_np

        bnch = self.calc_histogram(data, pct=self.pct, numbins=self.numbins)
        loval, hival = bnch.loval, bnch.hival
        return float(loval), float(hival)

    def calc_histogram(self, data, pct=1.0, numbins=2048):
        """Internal function used by this class."""

        self.logger.debug("Computing histogram, pct=%.4f numbins=%d" % (
            pct, numbins))
        height, width = data.shape[:2]
        self.logger.debug("Median analysis array is %dx%d" % (
            width, height))

        data = data[np.isfinite(data)]
        total_px = len(data.flat)
        if total_px == 0:
            return Bunch.Bunch(loval=0, hival=0)
        dist, bins = np.histogram(data, bins=numbins, density=False)

        cutoff = int((float(total_px) * (1.0 - pct)) / 2.0)
        top = len(dist) - 1
        self.logger.debug("top=%d cutoff=%d" % (top, cutoff))

        # calculate low cutoff
        cumsum = np.cumsum(dist)
        li = np.flatnonzero(cumsum > cutoff)
        if len(li) > 0:
            i = li[0]
            count_px = cumsum[i]
        else:
            i = 0
            count_px = 0
        if i > 0:
            nprev = cumsum[i - 1]
        else:
            nprev = 0
        loidx = i

        # interpolate between last two low bins
        val1, val2 = bins[i], bins[i + 1]
        divisor = float(count_px) - float(nprev)
        if divisor > 0.0:
            interp = (float(cutoff) - float(nprev)) / divisor
        else:
            interp = 0.0
        loval = val1 + ((val2 - val1) * interp)
        self.logger.debug("loval=%f val1=%f val2=%f interp=%f" % (
            loval, val1, val2, interp))

        # calculate high cutoff
        revdist = dist[::-1]
        cumsum = np.cumsum(revdist)
        li = np.flatnonzero(cumsum > cutoff)
        if len(li) > 0:
            i = li[0]
            count_px = cumsum[i]
        else:
            i = 0
            count_px = 0
        if i > 0:
            nprev = cumsum[i - 1]
        else:
            nprev = 0
        j = top - i
        hiidx = j + 1

        # interpolate between last two high bins
        val1, val2 = bins[j], bins[j + 1]
        divisor = float(count_px) - float(nprev)
        if divisor > 0.0:
            interp = (float(cutoff) - float(nprev)) / divisor
        else:
            interp = 0.0
        hival = val1 + ((val2 - val1) * interp)
        self.logger.debug("hival=%f val1=%f val2=%f interp=%f" % (
            hival, val1, val2, interp))

        return Bunch.Bunch(dist=dist, bins=bins, loval=loval, hival=hival,
                           loidx=loidx, hiidx=hiidx)


class StdDev(AutoCutsBase):
    """Calculate the cut levels based on a standard deviation analysis of
    the data.

    The calculation is:
        loval = hensa_lo * sdev(sample_data) + mean(sample_data)
        hival = hensa_hi * sdev(sample_data) + mean(sample_data)

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    sample : str (optional, 'crop', 'grid', or 'full', default: 'crop')
        Specifies how to access the image for calculation:
        - crop: crop an area from the middle of the image
        - grid: sample data in a grid pattern across the image
        - full: use the full image data

    full_px_limit : int (optional, defaults to 1M)
        Specifies the limit for using the full data if sample == 'full'.
        If the number of pixels in the image is larger than this, then
        the image will fall back to using a crop.

    num_points : int (optional, defaults to None)
        Specifies the number of points in the grid if sample == 'grid',
        or the diameter of the crop, if sample == 'crop' (or 'full' and
        number of pixels exceeds `full_px_limit`).  If None, the number
        pixels will be calculated to a "reasonable representative sample".

    hensa_lo : float (optional, defaults to -1.5)
        Specifies the low cut multiplication factor to apply to the
        standard deviation before adding the median (usually < 0)

    hensa_hi : float (optional, defaults to 4.0)
        Specifies the low cut multiplication factor to apply to the
        standard deviation before adding the median (usually > 0)

    """
    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='sample', type=str,
                  valid=['crop', 'grid', 'full'],
                  default='grid',
                  description="How to access data for calculation"),
            Param(name='full_px_limit', type=int,
                  default=1 * 1024 * 1024, allow_none=True,
                  description="For sample=full, fall back to crop if num_px > limit"),
            Param(name='num_points', type=int,
                  default=None, allow_none=True,
                  description="Number of points to sample (for sample=grid); 'None' for calculated default"),
            Param(name='hensa_lo', type=float, default=-1.5,
                  description="Low cut sdev multiplication factor"),
            Param(name='hensa_hi', type=float, default=4.0,
                  description="High cut sdev multiplication factor"),
        ]

    # NOTE: `usecrop` kwarg to be deprecated--accepted but not used
    # for backward compatibility with saved older settings
    def __init__(self, logger, usecrop=None, sample='grid',
                 full_px_limit=None, num_points=None,
                 hensa_lo=-1.5, hensa_hi=4.0):
        super().__init__(logger)

        self.kind = 'stddev'
        self.sample = sample
        self.full_px_limit = full_px_limit
        self.num_points = num_points
        # Constants used to calculate the lo and hi cut levels using the
        # "stddev" algorithm (from the old SOSS fits viewer)
        self.hensa_lo = hensa_lo
        self.hensa_hi = hensa_hi

    def calc_cut_levels(self, image):
        """See subclass documentation."""
        if self.sample == 'crop':
            if self.num_points is None:
                crop_radius = self.crop_radius
            else:
                crop_radius = int(self.num_points // 2)
            data = self.get_crop(image, crop_radius=crop_radius)
        elif self.sample == 'grid':
            data = self.get_sample(image, num_points=self.num_points)
        else:
            data = self.get_full(image, px_limit=self.full_px_limit)

        loval, hival = self.calc_stddev(data, hensa_lo=self.hensa_lo,
                                        hensa_hi=self.hensa_hi)
        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        """See subclass documentation."""
        if self.sample == 'crop':
            if self.num_points is None:
                crop_radius = self.crop_radius
            else:
                crop_radius = int(self.num_points // 2)
            data = self.get_crop_data(data_np, crop_radius=crop_radius)
        elif self.sample == 'grid':
            data = self.get_sample_data(data_np, num_points=self.num_points)
        else:
            data = data_np

        if data.size == 0:
            return (0, 0)
        loval, hival = self.calc_stddev(data, hensa_lo=self.hensa_lo,
                                        hensa_hi=self.hensa_hi)
        return float(loval), float(hival)

    def calc_stddev(self, data, hensa_lo=-1.5, hensa_hi=4.0):
        """Internal function used by this class."""
        data = data[np.isfinite(data)]
        if data.size == 0:
            return (0, 0)
        mean = np.mean(data)
        sdev = np.std(data)
        self.logger.debug(f"mean={mean} std={sdev}")

        loval = hensa_lo * sdev + mean
        hival = hensa_hi * sdev + mean

        return loval, hival


class MedianFilter(AutoCutsBase):
    """Calculate the cut levels based on a median filtering analysis of
    the data.

    The calculation is:
        out = median_filter(sample_data, size=length)
        loval, hival = min(out), max(out)

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    num_points : int (optional, defaults to None)
        Specifies the number of points to sample making up the grid.
        If None, the number of points will be calculated to a
        "reasonable representative sample".

    length : int (optional, defaults to 5)
        Specifies the size of the median filter to apply to the data

    """
    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='num_points', type=int,
                  default=2000, allow_none=True,
                  description="Number of points to sample; 'None' for calculated default"),
            Param(name='length', type=int, default=5,
                  description="Median kernel length"),
        ]

    def __init__(self, logger, num_points=2000, length=5):
        super().__init__(logger)

        self.kind = 'median'
        self.num_points = num_points
        self.length = length

    def calc_cut_levels(self, image):
        """See subclass documentation."""
        data = self.get_sample(image, num_points=self.num_points)

        loval, hival = self.calc_medianfilter(data, length=self.length)
        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        """See subclass documentation."""
        data = self.get_sample_data(data_np, num_points=self.num_points)

        loval, hival = self.calc_medianfilter(data, length=self.length)
        return float(loval), float(hival)

    def calc_medianfilter(self, data, length=5):
        """Internal function used by this class."""

        assert len(data.shape) >= 2, \
            AutoCutsError("input data should be 2D or greater")
        if length is None:
            length = 5

        xout = scipy.ndimage.filters.median_filter(data, size=length)
        loval = np.nanmin(xout)
        hival = np.nanmax(xout)

        return loval, hival


class ZScale(AutoCutsBase):
    """Calculate the cut levels based on a median filtering analysis of
    the data.

    Based on STScI's numdisplay implementation of IRAF's ZScale.

    The calculation is:
        local, hival = zscale(sample_data, contrast)

    Parameters
    ----------
    logger : :py:class:`~logging.Logger`
        Logger for tracing and debugging.

    contrast : float (optional, range: 0.0 - 1.0, defaults to 0.25)
        Specifies the contrast parameter to use in the zscale calculation

    num_points : int (optional, defaults to None)
        Specifies the number of points to sample making up the grid.
        If None, the number of points will be calculated to a
        "reasonable representative sample".

    """
    @classmethod
    def get_params_metadata(cls):
        return [
            Param(name='contrast', type=float,
                  default=0.25, allow_none=False,
                  description="Contrast"),
            Param(name='num_points', type=int,
                  default=1000, allow_none=True,
                  description="Number of points to sample; 'None' for calculated default"),
        ]

    def __init__(self, logger, contrast=0.25, num_points=None):
        super().__init__(logger)

        self.kind = 'zscale'
        self.contrast = contrast
        self.num_points = num_points

    def calc_cut_levels(self, image):
        """See subclass documentation."""
        data = self.get_sample(image, num_points=self.num_points)

        loval, hival = self.calc_zscale(data, contrast=self.contrast,
                                        num_points=self.num_points)
        return float(loval), float(hival)

    def calc_cut_levels_data(self, data_np):
        """See subclass documentation."""
        cutout = self.get_sample_data(data_np, num_points=self.num_points)

        loval, hival = self.calc_zscale(cutout, contrast=self.contrast,
                                        num_points=self.num_points)
        return float(loval), float(hival)

    def calc_zscale(self, data, contrast=0.25, num_points=1000):
        """Internal function used by this class."""
        # NOTE: num_per_row is ignored in this implementation

        assert len(data.shape) >= 2, \
            AutoCutsError("input data should be 2D or greater")
        ht, wd = data.shape[:2]

        # sanity check on contrast parameter
        assert (0.0 < contrast <= 1.0), \
            AutoCutsError("contrast (%.2f) not in range 0 < c <= 1" % (
                contrast))

        # remove masked elements, they cause problems
        data = data[np.logical_not(np.ma.getmaskarray(data))]
        # remove NaN and Inf from samples
        samples = data[np.isfinite(data)].flatten()
        samples = samples[:num_points]

        if samples.size == 0:
            return (0, 0)

        _interval = ZScaleInterval(samples.size, contrast=contrast)
        loval, hival = _interval.get_limits(samples)
        return loval, hival


# funky boolean converter
_bool = lambda st: str(st).lower() == 'true'  # noqa

autocuts_table = {
    'clip': Clip,
    'minmax': Minmax,
    'stddev': StdDev,
    'histogram': Histogram,
    'median': MedianFilter,
    'zscale': ZScale,
}


def get_autocuts(name):
    """Return the class object used to implement an autocuts algorithm.

    Parameters
    ----------
    name : str
        The name of the algorithm

    Returns
    -------
    klass : subclass of `~ginga.AutoCuts.AutoCutsBase`
            A class implementing the auto cut levels algorithm

    """
    if name not in autocut_methods:
        raise AutoCutsError("Method '%s' is not supported" % (name))

    return autocuts_table[name]


def get_autocuts_names():
    """Return the list of algorithm names for the available autocuts methods.

    Returns
    -------
    names : list of str
        A list of the known auto cut levels algorithm names

    """
    l = list(autocuts_table.keys())
    l.sort()
    return l



#! /usr/bin/env python
#
# example4_mpl.py -- Load a fits file into a Ginga widget with a
#          matplotlib backend.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
#
"""
   $ ./example4_mpl.py [fits file]

A Ginga object rendering to a generic matplotlib Figure.  In short,
this allows you to have all the interactive UI goodness of a Ginga widget
window in a matplotlib figure.  You can interactively flip, rotate, pan, zoom,
set cut levels and color map warp a FITS image.  Furthermore, you can plot
using matplotlib plotting on top of the image and the plots will follow all
the transformations.

See the Ginga quick reference
(http://ginga.readthedocs.io/en/latest/quickref.html)
for a list of the interactive features in the standard ginga widget.

example4 produces a simple matplotlib fits view with a couple of overplots.
This shows how you can use the functionality with straight python/matplotlib
sessions.  Run this by supplying a single FITS file on the command line.
"""

import sys

import matplotlib.pyplot as plt
import matplotlib.patches as patches

from ginga.mplw.ImageViewMpl import CanvasView
from ginga.misc import log
from ginga import cmap
from ginga.util.loader import load_data

# add matplotlib colormaps to ginga's own set
cmap.add_matplotlib_cmaps()

# Set to True to get diagnostic logging output
use_logger = True
logger = log.get_logger(null=not use_logger, log_stderr=True)

# create a regular matplotlib figure
fig = plt.figure()

# create a ginga object, initialize some defaults and
# tell it about the figure
fi = CanvasView(logger=logger)
fi.enable_autocuts('on')
fi.set_autocut_params('zscale')
fi.show_mode_indicator(True, corner='ur')
#fi.set_cmap(cmap.get_cmap('rainbow3'))
fi.set_figure(fig)
fi.show_mode_indicator(True, corner='ur')

# enable all interactive ginga features
fi.get_bindings().enable_all(True)

# load an image
if len(sys.argv) < 2:
    print("Please provide a FITS file on the command line")
    sys.exit(1)

image = load_data(sys.argv[1], logger=logger)
fi.set_image(image)
#fi.rotate(45)

# plot some example graphics via matplotlib

# Note adding axis from ginga (mpl backend) object
ax = fi.add_axes()

wd, ht = image.get_size()

# plot a line
l = ax.plot((wd * 0.33, wd * 0.75), (ht * 0.5, ht * 0.75), 'go-', c="g",
            label='line1')
# a rect
r = patches.Rectangle((wd * 0.10, ht * 0.10), wd * 0.6, ht * 0.5, ec='b',
                      fill=False)
ax.add_patch(r)

# if you rotate, flip, zoom or pan the the ginga image the graphics
# stay properly plotted.  See quickref of interactive ginga commands here:
#    http://ginga.readthedocs.io/en/latest/quickref.html
plt.show()



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.canvas.CanvasObject import get_canvas_types
from ginga import trcalc
from ginga.gw import Widgets

from .base import Stage, StageAction


class WhiteBalance(Stage):
    """
    White balance a 2D image.
    """

    _stagename = 'white-balance'

    def __init__(self):
        super().__init__()

        self.dc = get_canvas_types()
        self.crosshair_color = 'cyan'
        self._rval = 255
        self._gval = 255
        self._bval = 255
        self._lum = 255
        self.picker_obj = None
        self.viewer = None

        canvas = self.dc.DrawingCanvas()
        canvas.enable_edit(True)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.set_draw_mode('edit')
        self.canvas = canvas

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')

        fr = Widgets.Frame("White Balance")

        vbox = Widgets.VBox()
        captions = (('R:', 'label', 'rval', 'entry',
                     'G:', 'label', 'gval', 'entry',
                     'B:', 'label', 'bval', 'entry'),
                    ('Luminosity:', 'label', 'lum', 'entry')
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        # b.rval.set_limits(0, 255, 1)
        # b.gval.set_limits(0, 255, 1)
        # b.bval.set_limits(0, 255, 1)
        b.rval.set_text(str(self._rval))
        b.gval.set_text(str(self._gval))
        b.bval.set_text(str(self._bval))
        self._calc_lum()
        b.lum.set_text(str(self._lum))
        b.rval.add_callback('activated', self.chval_cb)
        b.gval.add_callback('activated', self.chval_cb)
        b.bval.add_callback('activated', self.chval_cb)
        b.lum.add_callback('activated', self.chlum_cb)

        vbox.add_widget(w, stretch=0)
        hbox = Widgets.HBox()
        hbox.set_spacing(2)
        self.w.white = Widgets.RadioButton("White")
        #self.w.gray = Widgets.RadioButton("Gray", group=self.w.white)
        self.w.white.set_state(True)
        self.w.white.set_enabled(False)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        hbox.add_widget(self.w.white, stretch=0)
        #hbox.add_widget(self.w.gray, stretch=0)
        vbox.add_widget(hbox, stretch=0)

        fr.set_widget(vbox)
        container.set_widget(fr)

        self.viewer = self.pipeline.get('viewer')
        self.canvas.set_surface(self.viewer)
        self.canvas.register_for_cursor_drawing(self.viewer)

        x, y = 0, 0
        self.picker_obj = self.dc.Crosshair(x, y, color=self.crosshair_color)

    @property
    def rval(self):
        return self._rval

    @rval.setter
    def rval(self, val):
        self._rval = val
        if self.gui_up:
            self.w.rval.set_text(str(self._rval))

    @property
    def gval(self):
        return self._gval

    @gval.setter
    def gval(self, val):
        self._gval = val
        if self.gui_up:
            self.w.gval.set_text(str(self._gval))

    @property
    def bval(self):
        return self._bval

    @bval.setter
    def bval(self, val):
        self._bval = val
        if self.gui_up:
            self.w.bval.set_text(str(self._bval))

    @property
    def lum(self):
        return self._lum

    @lum.setter
    def lum(self, val):
        self._lum = val
        if self.gui_up:
            self.w.lum.set_text("%.4f" % self._lum)

    def _calc_lum(self):
        #lum = (whiteR + whiteG + whiteB) / 3
        lum = (self._rval * 0.2126 + self._gval * 0.7152 + self._bval * 0.0722)
        self.lum = lum

    def edit_cb(self, canvas, picker_obj):
        old = self._get_state()

        x, y = picker_obj.points[0]
        x, y = int(x), int(y)
        #print(f"x, y is {x},{y}")

        idx = self.pipeline.index(self)
        data = self.pipeline.get_data(self.pipeline[idx - 1])
        self.verify_2d(data)
        tup = data[..., 0:3][y, x]
        self.rval, self.gval, self.bval = [int(n) for n in tup]
        self._calc_lum()
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="whbal / change val"))
        self.pipeline.run_from(self)

    def chval_cb(self, w):
        old = self._get_state()
        self.rval, self.gval, self.bval = [int(n)
                                           for n in [self.w.rval.get_text().strip(),
                                                     self.w.gval.get_text().strip(),
                                                     self.w.bval.get_text().strip()]]
        self._calc_lum()
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="whbal / change rgb"))
        self.pipeline.run_from(self)

    def chlum_cb(self, w):
        old = self._get_state()
        self.lum = float(self.w.lum.get_text().strip())
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="whbal / change lum"))
        self.pipeline.run_from(self)

    def _get_state(self):
        return dict(rval=self._rval, gval=self._gval, bval=self._bval,
                    lum=self._lum)

    def resume(self):
        # insert canvas, if not already
        p_canvas = self.viewer.get_canvas()
        if not p_canvas.has_object(self.canvas):
            p_canvas.add(self.canvas)

        if not self.canvas.has_object(self.picker_obj):
            self.canvas.add(self.picker_obj)

        self.canvas.ui_set_active(True, viewer=self.viewer)

    def pause(self):
        self.canvas.ui_set_active(False)

        # remove the canvas from the image
        p_canvas = self.viewer.get_canvas()
        try:
            p_canvas.delete_object(self.canvas)
        except Exception:
            pass

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        #imgR = imgR * lum / whiteR
        #imgG = imgG * lum / whiteG
        #imgB = imgB * lum / whiteB
        mn, mx = trcalc.get_minmax_dtype(data.dtype)
        res_np = data * self._lum / np.array((self._rval, self._gval,
                                              self._bval))
        res_np = res_np.clip(mn, mx).astype(data.dtype)
        print("calculated result")

        self.pipeline.send(res_np=res_np)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.rval = d['rval']
        self.gval = d['gval']
        self.bval = d['bval']
        self.lum = d['lum']



#
# CanvasRenderGL.py -- for rendering into a OpenGL widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time
import os.path
import numpy as np
import ctypes
import threading
from packaging.version import parse as parse_version

from OpenGL import GL as gl

from ginga.vec import CanvasRenderVec as vec
from ginga.canvas import render, transform
from ginga.cairow import CairoHelp
from ginga import trcalc, RGBMap
from ginga.util import rgb_cms

# Local imports
from .Camera import Camera
from . import GlHelp
from .glsl import __file__
shader_dir, _ = os.path.split(__file__)

# NOTE: we update the version later in gl_initialize()
opengl_version = parse_version('3.0')

# special scaling for GL text drawing to normalize it relative
# to other backends
_font_scale_factor = 2.0


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        # TODO: encapsulate this drawable
        self.cr = GlHelp.GlContext(surface)

        self._font_scale_factor = _font_scale_factor

    # def set_font(self, fontname, fontsize):
    #     fontsize = self.scale_fontsize(fontsize)
    #     self.font = self.cr.get_font(fontname, fontsize)

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        return self.cr.text_extents(text, font)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cp, cache, whence, order='RGB'):
        """Render the image represented by (rgb_arr) at (cx, cy)
        in the pixel space.
        """
        cp = np.asarray(cp, dtype=np.float32)

        # This bit is necessary because OpenGL assumes that pixels are
        # centered at 0.5 offsets from the start of the pixel.  Need to
        # follow the flip/swap transform of the viewer to make sure we
        # are applying the correct corner offset to each image corner
        # TODO: see if there is something we can set in OpenGL so that
        # we don't have to do this hack
        off = 0.5
        off = np.array(((-off, -off), (off, -off), (off, off), (-off, off)))
        tr = transform.FlipSwapTransform(self.viewer)
        cp += tr.to_(off)

        self.renderer.gl_draw_image(cvs_img, cp)

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        # TODO: this draws text as polygons, since there is no native
        # text support in OpenGL.  It uses cairo to convert the text to
        # paths.  Currently the paths are drawn, but not filled correctly.
        paths = CairoHelp.text_to_paths(text, font, flip_y=True,
                                        cx=cx, cy=cy, rot_deg=rot_deg)
        scale = self.viewer.get_scale()
        base = np.array((cx, cy))

        for pts in paths:
            # we have to rotate and scale the polygons to account for the
            # odd transform we use for OpenGL
            rot_deg = -self.viewer.get_rotation()
            if rot_deg != 0.0:
                pts = trcalc.rotate_coord(pts, [rot_deg], (cx, cy))
            pts = (pts - base) * (1 / scale) + base
            self.set_line(fill.color, alpha=fill.alpha, linewidth=1)
            # NOTE: since non-convex polygons are not filled correctly, it
            # doesn't work to set any fill here
            self.draw_polygon(pts, line=self.line)

    def draw_polygon(self, cpoints, line=None, fill=None):
        self.renderer.gl_draw_shape(gl.GL_LINE_LOOP, cpoints, fill, line)

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        # we have to approximate a circle in OpenGL
        # TODO: there is a more efficient algorithm described here:
        # http://slabode.exofire.net/circle_draw.shtml
        num_segments = 360
        cpoints = []
        for i in range(0, num_segments):
            theta = 2.0 * np.pi * i / float(num_segments)
            dx = cradius * np.cos(theta)
            dy = cradius * np.sin(theta)
            cpoints.append((cx + dx, cy + dy))

        self.renderer.gl_draw_shape(gl.GL_LINE_LOOP, cpoints, fill, line)

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        cpoints = [(cx1, cy1), (cx2, cy2)]
        self.renderer.gl_draw_shape(gl.GL_LINES, cpoints, None, line)

    def draw_path(self, cpoints, line=None):
        self.renderer.gl_draw_shape(gl.GL_LINE_STRIP, cpoints, None, line)


class CanvasRenderer(vec.VectorRenderMixin, render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)
        vec.VectorRenderMixin.__init__(self)

        self.kind = 'opengl'
        self.rgb_order = 'RGBA'
        self.surface = self.viewer.get_widget()
        self.use_offscreen_fbo = True

        # size of our GL viewport
        # these will change when the resize() is called
        self.wd, self.ht = 10, 10

        self.camera = Camera()
        self.camera.set_scene_radius(2)
        self.camera.set_camera_home_position((0, 0, 1000))
        self.camera.reset()

        self.mode3d = True
        self._drawing = False
        self._initialized = False
        self._tex_cache = dict()
        self._levels = (0.0, 0.0)
        self._cmap_len = 256
        self.max_texture_dim = 0
        self.image_uploads = []
        self.cmap_uploads = []

        self.pgm_mgr = GlHelp.ShaderManager(self.logger)

        self.fbo = None
        self.fbo_size = (0, 0)
        self.color_buf = None
        self.depth_buf = None
        self.lock = threading.RLock()

    def set_3dmode(self, tf):
        self.mode3d = tf
        self.gl_resize(self.wd, self.ht)
        scales = self.viewer.get_scale_xy()
        self.scale(scales)

    def _overlay_images(self, canvas, whence=0.0):
        #if not canvas.is_compound():
        if not hasattr(canvas, 'objects'):
            return

        for obj in canvas.get_objects():
            if hasattr(obj, 'prepare_image'):
                obj.prepare_image(self.viewer, whence)
            elif obj.is_compound() and (obj != canvas):
                self._overlay_images(obj, whence=whence)

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        if self._initialized:
            self._resize(dims)

            width, height = dims[:2]
            self.gl_resize(width, height)

            self.viewer.update_widget()

            # this is necessary for other widgets to get the same kind of
            # callback as for the standard pixel renderer
            self.viewer.make_callback('redraw', 0.0)

    def scale(self, scales):
        self.camera.scale_2d(scales[:2])

        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 0.0)

    def pan(self, pos):
        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 0.0)

    def rotate_2d(self, ang_deg):
        self.camera.rotate_2d(ang_deg)

        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 2.6)

    def rgbmap_change(self, rgbmap):
        if rgbmap not in self.cmap_uploads:
            self.cmap_uploads.append(rgbmap)
        #self.gl_set_cmap(rgbmap)

        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 2.0)

    def bg_change(self, bg):
        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 3.0)

    def levels_change(self, levels):
        self._levels = levels

        self.viewer.update_widget()
        # this is necessary for other widgets to get the same kind of
        # callback as for the standard pixel renderer
        self.viewer.make_callback('redraw', 1.0)

    def icc_profile_change(self):
        self.viewer.redraw(whence=0.1)

    def interpolation_change(self, interp):
        self.viewer.redraw(whence=0.0)

    def _common_draw(self, cvs_img, cache, whence):
        # internal common drawing phase for all images
        image = cvs_img.image
        if image is None:
            return
        viewer = self.viewer

        if (whence <= 0.0) or (cache.cutout is None) or (not cvs_img.optimize):

            # get destination location in data_coords
            dst_x, dst_y = cvs_img.crdmap.to_data((cvs_img.x, cvs_img.y))

            a1, b1, a2, b2 = 0, 0, image.width - 1, image.height - 1

            # scale by our scale
            _scale_x, _scale_y = cvs_img.scale_x, cvs_img.scale_y

            interp = cvs_img.interpolation
            if interp is None:
                t_ = viewer.get_settings()
                interp = t_.get('interpolation', 'basic')

            # previous choice might not be available if preferences
            # were saved when opencv was being used (and not used now);
            # if so, silently default to "basic"
            if interp not in trcalc.interpolation_methods:
                interp = 'basic'
            res = image.get_scaled_cutout2((a1, b1), (a2, b2),
                                           (_scale_x, _scale_y),
                                           method=interp)
            data = res.data

            if interp in ('basic', 'nearest'):
                cache.interp = 0
            elif interp in ('linear', 'bilinear'):
                cache.interp = 1
            elif interp in ('cubic', 'bicubic'):
                cache.interp = 2
            elif interp == 'lanczos':
                # TODO
                self.logger.warning("'lanczos' interpolation not yet implemented"
                                    " for this renderer--using 'bicubic' instead")
                #cache.interp = 3
                cache.interp = 2
            else:
                cache.interp = 0

            # We are limited by maximum texture size supported for the
            # OpenGl implementation.  Images larger than the maximum
            # in any dimension need to be downsampled to fit.
            ht, wd = data.shape[:2]
            extra = max(wd, ht) - self.max_texture_dim
            if extra > 0:
                new_wd, new_ht = wd - extra, ht - extra
                tup = trcalc.get_scaled_cutout_wdht(data, 0, 0, wd, ht,
                                                    new_wd, new_ht,
                                                    logger=self.logger)
                data = tup[0]

            if cvs_img.flipy:
                data = np.flipud(data)

            # calculate our offset
            pan_off = viewer.data_off
            cvs_x, cvs_y = dst_x - pan_off, dst_y - pan_off

            cache.cutout = data
            cache.cvs_pos = (cvs_x, cvs_y)

    def _prep_rgb_image(self, cvs_img, cache, whence):
        image = cvs_img.get_image()
        image_order = image.get_order()

        if whence <= 0.1 or cache.rgbarr is None:
            rgbarr = cache.cutout

            # convert to output ICC profile, if one is specified
            working_profile = rgb_cms.working_profile
            output_profile = self.viewer.t_.get('icc_output_profile', None)

            if working_profile is not None and output_profile is not None:
                if rgbarr is cache.cutout:
                    rgbarr = np.copy(rgbarr)
                self.convert_via_profile(rgbarr, image_order,
                                         working_profile, output_profile)
            else:
                rgbarr = self.reorder(self.rgb_order, rgbarr, image_order)

            ## depth = rgbarr.shape[2]
            ## if depth < 4:
            ##     # add an alpha channel if missing
            ##     _mn, _mx = trcalc.get_minmax_dtype(rgbarr.dtype)
            ##     rgbarr = trcalc.add_alpha(rgbarr, alpha=_mx)
            ## cache.image_type |= 0x2

            # array needs to be contiguous to transfer properly as buffer
            # to OpenGL
            cache.rgbarr = np.ascontiguousarray(rgbarr, dtype=rgbarr.dtype)

    def _prepare_image(self, cvs_img, cache, whence):
        self._common_draw(cvs_img, cache, whence)

        cache.image_type = 0x0     # no RGB mapping
        self._prep_rgb_image(cvs_img, cache, whence)

        if (whence <= 0.1) or (not cvs_img.optimize):
            # we can upload either 8 bpp or 16 bpp images
            if cache.rgbarr.dtype != np.dtype(np.uint8):
                cache.rgbarr = cache.rgbarr.astype(np.uint16)

        cache.drawn = True

    def _prepare_norm_image(self, cvs_img, cache, whence):
        t1 = t2 = t3 = t4 = time.time()

        self._common_draw(cvs_img, cache, whence)

        if cache.cutout is None:
            return

        t2 = time.time()
        if cvs_img.rgbmap is not None:
            rgbmap = cvs_img.rgbmap
        else:
            rgbmap = self.viewer.get_rgbmap()

        cache.image_type = 0x1     # use RGB mapping
        image = cvs_img.get_image()
        image_order = image.get_order()

        if (whence <= 0.1) or (cache.rgbarr is None) or (not cvs_img.optimize):

            img_arr = cache.cutout

            if len(img_arr.shape) == 2:
                # <-- monochrome image with no alpha
                cache.rgbarr = img_arr.astype(np.float32)

            elif img_arr.shape[2] < 3:
                # <-- monochrome image with alpha
                cache.image_type |= 0x2
                cache.rgbarr = img_arr.astype(np.float32)

            else:
                # <-- RGB[A] image
                self._prep_rgb_image(cvs_img, cache, whence)
                cache.image_type |= 0x4
                cache.rgbarr = cache.rgbarr.astype(np.float32)

        cache.drawn = True
        t5 = time.time()
        self.logger.debug("draw: t2=%.4f t3=%.4f t4=%.4f t5=%.4f total=%.4f" % (
            t2 - t1, t3 - t2, t4 - t3, t5 - t4, t5 - t1))

    def prepare_image(self, cvs_img, cache, whence):
        if cvs_img.kind == 'image':
            self._prepare_image(cvs_img, cache, whence)

        elif cvs_img.kind == 'normimage':
            self._prepare_norm_image(cvs_img, cache, whence)

        else:
            raise render.RenderError("I don't know how to render canvas type '{}'".format(cvs_img.kind))

        img_arr = cache.rgbarr
        self.image_uploads.append((cvs_img.image_id, img_arr, cache.image_type))

    def convert_via_profile(self, data_np, order, inprof_name, outprof_name):
        """Convert the given RGB data from the working ICC profile
        to the output profile in-place.

        Parameters
        ----------
        data_np : ndarray
            RGB image data to be displayed.

        order : str
            Order of channels in the data (e.g. "BGRA").

        inprof_name, outprof_name : str
            ICC profile names (see :func:`ginga.util.rgb_cms.get_profiles`).

        """
        t_ = self.viewer.get_settings()
        # get rest of necessary conversion parameters
        to_intent = t_.get('icc_output_intent', 'perceptual')
        proofprof_name = t_.get('icc_proof_profile', None)
        proof_intent = t_.get('icc_proof_intent', 'perceptual')
        use_black_pt = t_.get('icc_black_point_compensation', False)

        try:
            rgbobj = RGBMap.RGBPlanes(data_np, order)
            arr_np = rgbobj.get_array('RGB')

            arr = rgb_cms.convert_profile_fromto(arr_np, inprof_name, outprof_name,
                                                 to_intent=to_intent,
                                                 proof_name=proofprof_name,
                                                 proof_intent=proof_intent,
                                                 use_black_pt=use_black_pt,
                                                 logger=self.logger)
            ri, gi, bi = rgbobj.get_order_indexes('RGB')

            out = data_np
            out[..., ri] = arr[..., 0]
            out[..., gi] = arr[..., 1]
            out[..., bi] = arr[..., 2]

            self.logger.debug("Converted from '%s' to '%s' profile" % (
                inprof_name, outprof_name))

        except Exception as e:
            self.logger.warning("Error converting output from working profile: %s" % (str(e)))
            # TODO: maybe should have a traceback here
            self.logger.info("Output left unprofiled")

    def get_texture_id(self, image_id):
        tex_id = self._tex_cache.get(image_id, None)
        if tex_id is None:
            context = self.viewer.make_context_current()
            tex_id = gl.glGenTextures(1)
            self._tex_cache[image_id] = tex_id
        return tex_id

    def render_whence(self, whence):
        if whence <= 2.0:
            p_canvas = self.viewer.get_private_canvas()
            self._overlay_images(p_canvas, whence=whence)

    def get_camera(self):
        return self.camera

    def getOpenGLInfo(self):
        self.max_texture_dim = gl.glGetIntegerv(gl.GL_MAX_TEXTURE_SIZE)
        info = dict(vendor=gl.glGetString(gl.GL_VENDOR).decode(),
                    renderer=gl.glGetString(gl.GL_RENDERER).decode(),
                    opengl_version=gl.glGetString(gl.GL_VERSION).decode(),
                    shader_version=gl.glGetString(gl.GL_SHADING_LANGUAGE_VERSION).decode(),
                    max_tex="{}x{}".format(self.max_texture_dim,
                                           self.max_texture_dim))
        return info

    def gl_initialize(self):
        global opengl_version
        context = self.viewer.make_context_current()

        d = self.getOpenGLInfo()
        self.logger.info("OpenGL info--Vendor: '%(vendor)s'  "
                         "Renderer: '%(renderer)s'  "
                         "Version: '%(opengl_version)s'  "
                         "Shader: '%(shader_version)s' "
                         "Max texture: '%(max_tex)s'" % d)

        opengl_version = parse_version(d['opengl_version'].split(' ')[0])

        if self.use_offscreen_fbo:
            self.create_offscreen_fbo()

        # --- line drawing shaders ---
        self.pgm_mgr.load_program('shape', shader_dir)

        # --- setup VAO for line drawing ---
        shader = self.pgm_mgr.setup_program('shape')

        # Create a new VAO (Vertex Array Object) and bind it
        self.vao_line = gl.glGenVertexArrays(1)
        gl.glBindVertexArray(self.vao_line)

        # Generate buffers to hold our vertices
        self.vbo_line = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vbo_line)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, None, gl.GL_DYNAMIC_DRAW)
        # Get the position of the 'position' in parameter of our shader
        # and bind it.
        _pos = gl.glGetAttribLocation(shader, 'position')
        gl.glEnableVertexAttribArray(_pos)
        # Describe the position data layout in the buffer
        gl.glVertexAttribPointer(_pos, 3, gl.GL_FLOAT, False, 0,
                                 ctypes.c_void_p(0))

        # Unbind the VAO first (important)
        gl.glBindVertexArray(0)
        #gl.glDisableVertexAttribArray(_pos)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        self.pgm_mgr.setup_program(None)

        # --- image drawing shaders ---
        self.pgm_mgr.load_program('image', shader_dir)
        shader = self.pgm_mgr.setup_program('image')

        # --- setup VAO for image drawing ---
        self.vao_img = gl.glGenVertexArrays(1)
        gl.glBindVertexArray(self.vao_img)

        # Generate buffers to hold our vertices
        self.vbo_img = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vbo_img)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, None, gl.GL_DYNAMIC_DRAW)
        # Get the position of the 'position' in parameter of our shader
        # and bind it.
        _pos = gl.glGetAttribLocation(shader, 'position')
        gl.glEnableVertexAttribArray(_pos)
        # Describe the position data layout in the buffer
        gl.glVertexAttribPointer(_pos, 3, gl.GL_FLOAT, False, 5 * 4,
                                 ctypes.c_void_p(0))
        _pos2 = gl.glGetAttribLocation(shader, 'i_tex_coord')
        gl.glEnableVertexAttribArray(_pos2)
        gl.glVertexAttribPointer(_pos2, 2, gl.GL_FLOAT, False, 5 * 4,
                                 ctypes.c_void_p(3 * 4))

        # Unbind the VAO first
        gl.glBindVertexArray(0)
        #gl.glDisableVertexAttribArray(_pos)
        #gl.glDisableVertexAttribArray(_pos2)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        self.pgm_mgr.setup_program(None)

        self.cmap_buf = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, self.cmap_buf)
        gl.glBufferData(gl.GL_TEXTURE_BUFFER, self._cmap_len * 4, None,
                        gl.GL_DYNAMIC_DRAW)
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, 0)

        rgbmap = self.viewer.get_rgbmap()
        if rgbmap not in self.cmap_uploads:
            self.cmap_uploads.append(rgbmap)

        gl.glDisable(gl.GL_CULL_FACE)
        gl.glFrontFace(gl.GL_CCW)
        self._initialized = True

    def gl_set_image(self, tex_id, img_arr, image_type):
        """NOTE: this is a slow operation--downloading a texture."""
        #context = self.viewer.make_context_current()

        ht, wd = img_arr.shape[:2]

        gl.glBindTexture(gl.GL_TEXTURE_2D, tex_id)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
                           gl.GL_NEAREST)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER,
                           gl.GL_NEAREST)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_R,
                           gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S,
                           gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T,
                           gl.GL_CLAMP_TO_EDGE)

        # see image_type in image fragment shader
        if image_type & 0x1 == 0:
            # "native" image colors--3 color image, no RGBMAP
            if img_arr.dtype == np.dtype(np.uint8):
                if img_arr.shape[2] == 3:
                    # 8bpp RGB
                    gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
                    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB, wd, ht, 0,
                                    gl.GL_RGB, gl.GL_UNSIGNED_BYTE, img_arr)
                    self.logger.debug("uploaded 8 bpp RGB as texture {}".format(tex_id))
                elif img_arr.shape[2] == 4:
                    # 8bpp RGBA
                    gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
                    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, wd, ht, 0,
                                    gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, img_arr)
                    self.logger.debug("uploaded 8 bpp RGBA as texture {}".format(tex_id))
            elif img_arr.dtype == np.dtype(np.uint16):
                if img_arr.shape[2] == 3:
                    # 16bpp RGB
                    gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 2)
                    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB16, wd, ht, 0,
                                    gl.GL_RGB, gl.GL_UNSIGNED_SHORT, img_arr)
                    self.logger.debug("uploaded 16 bpp RGB as texture {}".format(tex_id))
                elif img_arr.shape[2] == 4:
                    # 16bpp RGBA
                    gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 2)
                    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA16, wd, ht, 0,
                                    gl.GL_RGBA, gl.GL_UNSIGNED_SHORT, img_arr)
                    self.logger.debug("uploaded 16 bpp RGBA as texture {}".format(tex_id))
            else:
                raise ValueError("unknown image type: {}".format(hex(image_type)))

            self.logger.debug("uploaded rgbarr as texture {}".format(tex_id))

        else:
            if len(img_arr.shape) < 3 or img_arr.shape[2] == 1:
                # mono, no alpha
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_R32F, wd, ht, 0,
                                gl.GL_RED, gl.GL_FLOAT, img_arr)
                self.logger.debug("uploaded mono as texture {}".format(tex_id))
            elif len(img_arr.shape) == 3 and img_arr.shape[2] == 2:
                # mono, with alpha
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RG32F, wd, ht, 0,
                                gl.GL_RG, gl.GL_FLOAT, img_arr)
                self.logger.debug("uploaded mono w/alpha as texture {}".format(tex_id))
            elif len(img_arr.shape) == 3 and img_arr.shape[2] == 3:
                # RGB, no alpha
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB32F, wd, ht, 0,
                                gl.GL_RGB, gl.GL_FLOAT, img_arr)
                self.logger.debug("uploaded RGB as texture {}".format(tex_id))
            elif len(img_arr.shape) == 3 and img_arr.shape[2] == 4:
                # RGBA
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA32F, wd, ht, 0,
                                gl.GL_RGBA, gl.GL_FLOAT, img_arr)
                self.logger.debug("uploaded RGBA as texture {}".format(tex_id))

            else:
                raise ValueError("unknown image type: {}".format(hex(image_type)))

        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

    def gl_set_cmap(self, rgbmap):
        # TODO: this does not yet work with 'histeq' color distribution
        hashsize = rgbmap.get_hash_size()
        idx = rgbmap.get_hasharray(np.arange(0, hashsize))
        if hashsize != self._cmap_len:
            # Downsample color distribution hash to our opengl colormap length
            xi = (np.arange(0, self._cmap_len) * (hashsize / self._cmap_len)).clip(0, hashsize).astype(np.uint)
            if len(xi) != self._cmap_len:
                raise render.RenderError("Error generating color hash table index: size mismatch {} != {}".format(len(xi), self._cmap_len))

            idx = idx[xi]

        img_arr = np.ascontiguousarray(rgbmap.get_colors(),
                                       dtype=np.uint8)
        img_arr = img_arr[idx]

        # append alpha channel
        wd = img_arr.shape[0]
        alpha = np.full((wd, 1), self._cmap_len - 1, dtype=np.uint8)
        img_arr = np.concatenate((img_arr, alpha), axis=1)
        map_id = self.get_texture_id(rgbmap.mapper_id)

        # transfer colormap info to GPU buffer
        #context = self.viewer.make_context_current()
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, self.cmap_buf)
        gl.glBufferSubData(gl.GL_TEXTURE_BUFFER, 0, img_arr)
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, 0)
        self.logger.debug("uploaded cmap as texture buffer {}".format(map_id))

    def gl_draw_image(self, cvs_img, cp):
        if not self._drawing:
            # this test ensures that we are not trying to draw before
            # the OpenGL context is set for us correctly
            return

        cache = cvs_img.get_cache(self.viewer)
        # TODO: put tex_id in cache?
        tex_id = self.get_texture_id(cvs_img.image_id)
        rgbmap = self.viewer.get_rgbmap()
        map_id = self.get_texture_id(rgbmap.mapper_id)
        self.pgm_mgr.setup_program('image')
        gl.glBindVertexArray(self.vao_img)

        _loc = self.pgm_mgr.get_uniform_loc("projection")
        gl.glUniformMatrix4fv(_loc, 1, False, self.camera.proj_mtx)
        _loc = self.pgm_mgr.get_uniform_loc("view")
        gl.glUniformMatrix4fv(_loc, 1, False, self.camera.view_mtx)

        gl.glActiveTexture(gl.GL_TEXTURE0 + 0)
        gl.glBindTexture(gl.GL_TEXTURE_2D, tex_id)
        _loc = self.pgm_mgr.get_uniform_loc("img_texture")
        gl.glUniform1i(_loc, 0)

        gl.glActiveTexture(gl.GL_TEXTURE0 + 1)
        gl.glBindTexture(gl.GL_TEXTURE_BUFFER, map_id)
        gl.glTexBuffer(gl.GL_TEXTURE_BUFFER, gl.GL_RGBA8UI, self.cmap_buf)
        _loc = self.pgm_mgr.get_uniform_loc("color_map")
        gl.glUniform1i(_loc, 1)

        _loc = self.pgm_mgr.get_uniform_loc("image_type")
        gl.glUniform1i(_loc, cache.image_type)

        _loc = self.pgm_mgr.get_uniform_loc("interp")
        gl.glUniform1i(_loc, cache.interp)

        # if image has fixed cut levels, use those
        cuts = getattr(cvs_img, 'cuts', None)
        if cuts is not None:
            loval, hival = cuts
        else:
            loval, hival = self._levels

        _loc = self.pgm_mgr.get_uniform_loc("loval")
        gl.glUniform1f(_loc, loval)

        _loc = self.pgm_mgr.get_uniform_loc("hival")
        gl.glUniform1f(_loc, hival)

        # pad with z=0 coordinate if lacking
        vertices = trcalc.pad_z(cp, dtype=np.float32)

        # Send the data over to the buffer
        texcoord = np.array([(0.0, 0.0), (1.0, 0.0),
                             (1.0, 1.0), (0.0, 1.0)], dtype=np.float32)
        data = np.concatenate((vertices, texcoord), axis=1)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vbo_img)
        # see https://www.khronos.org/opengl/wiki/Buffer_Object_Streaming
        #gl.glBufferData(gl.GL_ARRAY_BUFFER, None, gl.GL_DYNAMIC_DRAW)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, data, gl.GL_DYNAMIC_DRAW)

        gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

        gl.glDrawArrays(gl.GL_TRIANGLE_FAN, 0, 4)

        gl.glBindVertexArray(0)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        self.pgm_mgr.setup_program(None)

    def gl_draw_shape(self, gl_shape, cpoints, fill, line):

        if not self._drawing:
            # this test ensures that we are not trying to draw before
            # the OpenGL context is set for us correctly
            return

        # pad with z=0 coordinate if lacking
        z_pts = trcalc.pad_z(cpoints, dtype=np.float32)

        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        self.pgm_mgr.setup_program('shape')
        gl.glBindVertexArray(self.vao_line)

        # Update the vertices data in the VBO
        vertices = z_pts.astype(np.float32)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vbo_line)
        # see https://www.khronos.org/opengl/wiki/Buffer_Object_Streaming
        #gl.glBufferData(gl.GL_ARRAY_BUFFER, None, gl.GL_DYNAMIC_DRAW)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, vertices, gl.GL_DYNAMIC_DRAW)

        _loc = self.pgm_mgr.get_uniform_loc("projection")
        gl.glUniformMatrix4fv(_loc, 1, False, self.camera.proj_mtx)
        _loc = self.pgm_mgr.get_uniform_loc("view")
        gl.glUniformMatrix4fv(_loc, 1, False, self.camera.view_mtx)

        # update color uniform
        _loc = self.pgm_mgr.get_uniform_loc("fg_clr")

        # draw fill, if any
        if fill is not None and fill.color is not None:
            _c = fill._color_4tup
            gl.glUniform4f(_loc, _c[0], _c[1], _c[2], _c[3])

            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

            # TODO: this will not fill in non-convex polygons correctly
            gl.glDrawArrays(gl.GL_TRIANGLE_FAN, 0, len(vertices))

        # draw line, if any
        # TODO: support line stippling (dash)
        if line is not None and line.linewidth > 0:
            _c = line._color_4tup
            gl.glUniform4f(_loc, _c[0], _c[1], _c[2], _c[3])

            # draw outline
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            # > 1.0 not guaranteed to be supported as of OpenGL 4.2
            # TODO
            # gl.glLineWidth(line.linewidth)
            gl.glLineWidth(1.0)

            gl.glDrawArrays(gl_shape, 0, len(vertices))

        gl.glBindVertexArray(0)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
        self.pgm_mgr.setup_program(None)

    def show_errors(self):
        while True:
            err = gl.glGetError()
            if err == gl.GL_NO_ERROR:
                return
            self.logger.error("gl error: {}".format(err))

    def gl_resize(self, width, height):
        self.wd, self.ht = width, height

        context = self.viewer.make_context_current()

        gl.glViewport(0, 0, width, height)

        self.camera.set_viewport_dimensions(width, height)
        self.camera.calc_gl_transform()

    def gl_paint(self):
        with self.lock:
            context = self.viewer.make_context_current()

            # perform any necessary image updates
            uploads, self.image_uploads = self.image_uploads, []
            for image_id, img_arr, image_type in uploads:
                tex_id = self.get_texture_id(image_id)
                self.gl_set_image(tex_id, img_arr, image_type)

            # perform any necessary rgbmap updates
            rgbmap = self.viewer.get_rgbmap()
            if rgbmap not in self.cmap_uploads:
                self.cmap_uploads.append(rgbmap)
            uploads, self.cmap_uploads = self.cmap_uploads, []
            for rgbmap in uploads:
                self.gl_set_cmap(rgbmap)

            self._drawing = True
            try:
                gl.glDepthFunc(gl.GL_LEQUAL)
                gl.glEnable(gl.GL_DEPTH_TEST)

                r, g, b = self.viewer.img_bg
                gl.glClearColor(r, g, b, 1.0)
                gl.glClearDepth(1.0)
                gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

                cr = RenderContext(self, self.viewer, self.surface)
                self.draw_vector(cr)

            finally:
                self._drawing = False
                gl.glFlush()
                self.show_errors()

    def create_offscreen_fbo(self):
        if self.fbo is not None:
            self.delete_fbo_buffers()
        width, height = self.dims
        self.fbo_size = self.dims
        self.color_buf = gl.glGenRenderbuffers(1)
        self.depth_buf = gl.glGenRenderbuffers(1)

        # binds created FBO to context both for read and draw
        self.fbo = gl.glGenFramebuffers(1)
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self.fbo)
        gl.glViewport(0, 0, width, height)

        # bind color render buffer
        gl.glBindRenderbuffer(gl.GL_RENDERBUFFER, self.color_buf)
        gl.glRenderbufferStorage(gl.GL_RENDERBUFFER, gl.GL_RGBA8, width, height)
        gl.glFramebufferRenderbuffer(gl.GL_FRAMEBUFFER, gl.GL_COLOR_ATTACHMENT0,
                                     gl.GL_RENDERBUFFER, self.color_buf)

        # bind depth render buffer
        ## gl.glBindRenderbuffer(gl.GL_RENDERBUFFER, self.depth_buf)
        ## gl.glRenderbufferStorage(gl.GL_RENDERBUFFER, gl.GL_DEPTH_COMPONENT,
        ##                          width, height)
        ## gl.glFramebufferRenderbuffer(gl.GL_FRAMEBUFFER, gl.GL_DEPTH_ATTACHMENT,
        ##                              gl.GL_RENDERBUFFER, self.depth_buf)

        self.drawbuffers = [gl.GL_COLOR_ATTACHMENT0]
        gl.glDrawBuffers(1, self.drawbuffers)

        # check FBO status
        # TODO: returning a non-zero status, even though it seems to be working
        # fine.
        ## status = gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER)
        ## if status != gl.GL_FRAMEBUFFER_COMPLETE:
        ##     raise render.RenderError("Error initializing offscreen framebuffer: status={}".format(status))
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)

    def delete_fbo_buffers(self):
        if self.color_buf is not None:
            gl.glDeleteRenderbuffers(1, [self.color_buf])
            self.color_buf = None
        if self.depth_buf is not None:
            gl.glDeleteRenderbuffers(1, [self.depth_buf])
            self.depth_buf = None
        if self.fbo is not None:
            gl.glDeleteFramebuffers(1, [self.fbo])
            self.fbo = None
        self.fbo_size = (0, 0)

    def get_surface_as_array(self, order='RGBA'):
        if self.dims != self.fbo_size:
            self.create_offscreen_fbo()
        width, height = self.dims

        context = self.viewer.make_context_current()

        # some widget sets use a non-default FBO for rendering, so save
        # and restore
        cur_fbo = gl.glGetIntegerv(gl.GL_FRAMEBUFFER_BINDING)

        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self.fbo)
        gl.glViewport(0, 0, width, height)
        ## if self.use_offscreen_fbo:
        ##     gl.glReadBuffer(gl.GL_COLOR_ATTACHMENT0)

        try:
            self.gl_paint()
            img_buf = gl.glReadPixels(0, 0, width, height, gl.GL_RGBA,
                                      gl.GL_UNSIGNED_BYTE)
        finally:
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, cur_fbo)

        # seems to be necessary to redraw the main window
        self.viewer.update_widget()

        img_np = np.frombuffer(img_buf, dtype=np.uint8).reshape(height,
                                                                width, 4)
        img_np = np.flipud(img_np)

        if order is None or order == 'RGBA':
            return img_np

        img_np = trcalc.reorder_image(order, img_np, 'RGBA')
        return img_np

    def initialize(self):
        self.rl = []

    def finalize(self):
        # for this renderer, this is handled in gl_paint()
        pass

    def setup_cr(self, shape):
        # special cr that just stores up a render list
        cr = vec.RenderContext(self, self.viewer, self.surface)
        cr._font_scale_factor = _font_scale_factor
        cr.initialize_from_shape(shape, font=False)
        return cr

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.surface)
        return cr.text_extents(text, font=font)

    def calc_const_len(self, clen):
        # zoom is accomplished by viewing distance in OpenGL, so we
        # have to adjust clen by scale to get a constant size
        scale = self.viewer.get_scale_max()
        return clen / scale

    def scale_fontsize(self, fontsize):
        return fontsize

    def get_dimensions(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        font = cr.get_font_from_shape(shape)
        return cr.text_extents(shape.text, font=font)



#
# rvmode.py -- special bindings for the reference viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.modes.mode_base import Mode


class RVMode(Mode):
    """Reference Viewer Mode enables bindings that are specific to use in
    the Ginga Reference image viewer, and not in standalone programs using
    the viewer widget.

    Default bindings in mode
    ------------------------
    Z : raise Zoom plugin (if activated)
    I : raise Info plugin (if activated)
    H : raise Header plugin (if activated)
    C : raise Contents plugin (if activated)
    D : raise Dialogs tab
    h : show help message for the current mode
    F : go borderless fullscreen
    f : toggle fullscreen
    m : toggle maximized
    < : toggle collapse left pane
    > : toggle collapse right pane
    up arrow : previous image in channel
    down arrow : next image in channel
    J : cycle workspace type
    k : add channel auto
    K : remove current channel
    f1 : show channel names
    left arrow : previous channel in workspace
    right arrow : next channel in workspace

    motion : show info under the cursor
    left click : focus this viewer/channel
    """
    # Needs to be set by reference viewer (via set_shell_ref) before any
    # channel viewers are created
    fv = None

    @classmethod
    def set_shell_ref(cls, fv):
        cls.fv = fv

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_rvmode=['__m', None, None],

            kp_raise_zoom=['Z'],
            kp_raise_info=['I'],
            kp_raise_header=['H'],
            kp_raise_contents=['C'],
            kp_raise_dialogs=['D'],
            kp_show_mode_help=['h'],

            kp_go_fullscreen=['F'],
            kp_toggle_fullscreen=['f'],
            kp_toggle_maximize=['m'],

            kp_collapse_pane_left=['<'],
            kp_collapse_pane_right=['>'],

            # channel and workspace
            kp_previous_image_in_channel=['up'],
            kp_next_image_in_channel=['down'],
            kp_cycle_workspace_type=['J'],
            kp_add_channel_auto=['k'],
            kp_remove_channel_auto=['K'],
            kp_show_channel_names=['f1'],
            kp_previous_channel_in_workspace=['left'],
            kp_next_channel_in_workspace=['n', 'right'])

        viewer.set_callback('none-move', self.ms_showxy)

    def __str__(self):
        return 'rvmode'

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_raise_zoom(self, viewer, event, data_x, data_y, msg=True):
        self.fv.ds.raise_tab('Zoom')
        return True

    def kp_raise_info(self, viewer, event, data_x, data_y, msg=True):
        self.fv.ds.raise_tab('Info')
        return True

    def kp_raise_header(self, viewer, event, data_x, data_y, msg=True):
        self.fv.ds.raise_tab('Header')
        return True

    def kp_raise_contents(self, viewer, event, data_x, data_y, msg=True):
        self.fv.ds.raise_tab('Contents')
        return True

    def kp_raise_dialogs(self, viewer, event, data_x, data_y, msg=True):
        self.fv.ds.raise_tab('Dialogs')
        return True

    def kp_go_fullscreen(self, viewer, event, data_x, data_y, msg=True):
        self.fv.build_fullscreen()
        return True

    def kp_toggle_fullscreen(self, viewer, event, data_x, data_y, msg=True):
        self.fv.toggle_fullscreen()
        return True

    def kp_toggle_maximize(self, viewer, event, data_x, data_y, msg=True):
        self.fv.maximize()
        return True

    def kp_collapse_pane_left(self, viewer, event, data_x, data_y, msg=True):
        self.fv.collapse_pane('left')
        return True

    def kp_collapse_pane_right(self, viewer, event, data_x, data_y, msg=True):
        self.fv.collapse_pane('right')
        return True

    def kp_cycle_workspace_type(self, viewer, event, data_x, data_y, msg=True):
        self.fv.cycle_workspace_type()
        return True

    def kp_add_channel_auto(self, viewer, event, data_x, data_y, msg=True):
        self.fv.add_channel_auto()
        return True

    def kp_remove_channel_auto(self, viewer, event, data_x, data_y, msg=True):
        self.fv.remove_channel_auto()
        return True

    def kp_show_channel_names(self, viewer, event, data_x, data_y, msg=True):
        self.fv.show_channel_names()
        return True

    def kp_previous_image_in_channel(self, viewer, event, data_x, data_y,
                                     msg=True):
        self.fv.prev_img()
        return True

    def kp_next_image_in_channel(self, viewer, event, data_x, data_y, msg=True):
        self.fv.next_img()
        return True

    def kp_previous_channel_in_workspace(self, viewer, event, data_x, data_y,
                                         msg=True):
        self.fv.prev_channel()
        return True

    def kp_next_channel_in_workspace(self, viewer, event, data_x, data_y,
                                     msg=True):
        self.fv.next_channel()
        return True

    def kp_show_mode_help(self, viewer, event, data_x, data_y, msg=True):
        bm = viewer.get_bindmap()
        mode_name, mode_type = bm.current_mode()
        if mode_name is not None:
            bd = viewer.get_bindings()
            mode_obj = bd.get_mode_obj(mode_name)
            name, docstr = mode_obj.get_docstring()
            self.fv.show_help_text(name, docstr, wsname='right')
        return True

    #####  SCROLL ACTION CALLBACKS #####

    #####  MOUSE ACTION CALLBACKS #####

    def ms_showxy(self, viewer, event, data_x, data_y):
        """Motion event in the channel viewer window.  Show the pointing
        information under the cursor.
        """
        self.fv.showxy(viewer, data_x, data_y)
        return False

    # def ms_focus_viewer(self, viewer, event, data_x, data_y):
    #     if event.state == 'down':
    #         chname = self.fv.get_channel_name(viewer)
    #         if chname is not None:
    #             self.fv.force_focus(chname)
    #     # NOTE: this return of False is important!
    #     return False



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path

import numpy as np
from PIL import Image
from astropy.io import fits

from ginga.gw import Widgets
from ginga.util.paths import ginga_home
from ginga.util.action import AttrAction

from .base import Stage


class Output(Stage):

    _stagename = 'output'

    def __init__(self):
        super().__init__()

        self.fv = None
        self.in_image = None
        self._output_folder = '.'
        self._output_filetype = "PNG"
        self._auto_save = False
        self.output_types = ["FITS", "PNG", "JPEG"]
        self.convert_info = dict(FITS=dict(format="fits", ext=".fits"),
                                 PNG=dict(format="png", ext=".png"),
                                 JPEG=dict(format="jpeg", ext=".jpg"))

    def build_gui(self, container):
        self.fv = self.pipeline.get("fv")

        fr = Widgets.Frame("Output")

        captions = [('Output Folder:', 'label', 'output_folder', 'entryset'),
                    ('Output Filename:', 'label', 'output_filename', 'entry'),
                    ('Output Filetype:', 'label', 'output_filetype', 'combobox'),
                    ('Save', 'button', 'Auto Save', 'checkbox'),
                    ]
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        b.output_folder.set_text(self._output_folder)
        b.output_folder.set_tooltip("Folder for pipeline output")
        b.output_folder.add_callback('activated', self.set_folder_cb)
        b.output_filename.set_text('')
        b.output_filename.set_tooltip("File for pipeline output")
        #b.output_filename.add_callback('activated', self.set_filename_cb)
        for name in self.output_types:
            b.output_filetype.append_text(name)
        idx = self.output_types.index(self._output_filetype)
        b.output_filetype.set_index(idx)
        b.output_filetype.add_callback('activated', self.set_filetype_cb)
        b.save.add_callback('activated', self.save_as_cb)
        b.auto_save.set_state(self._auto_save)
        b.auto_save.set_tooltip("Automatically save file when clicked")
        b.auto_save.add_callback('activated', self.autosave_cb)

        fr.set_widget(w)
        container.set_widget(fr)

    @property
    def output_folder(self):
        return self._output_folder

    @output_folder.setter
    def output_folder(self, val):
        self._output_folder = val
        if self.gui_up:
            self.w.output_folder.set_text(val)

    @property
    def output_filetype(self):
        return self._output_filetype

    @output_filetype.setter
    def output_filetype(self, val):
        self._output_filetype = val
        if self.gui_up:
            idx = self.output_types.index(self._output_filetype)
            self.w.output_filetype.set_index(idx)

    @property
    def auto_save(self):
        return self._auto_save

    @auto_save.setter
    def auto_save(self, val):
        self._auto_save = val
        if self.gui_up:
            self.w.auto_save.set_state(val)

    def set_folder_cb(self, widget):
        old = dict(output_folder=self._output_folder)
        folder = widget.get_text().strip()
        self.output_folder = folder
        new = dict(output_folder=self._output_folder)
        self.pipeline.push(AttrAction(self, old, new,
                                      descr="output / change output folder"))

    def set_filetype_cb(self, widget, idx):
        old = dict(output_filetype=self._output_filetype)
        self._output_filetype = widget.get_text()
        new = dict(output_filetype=self._output_filetype)
        self.pipeline.push(AttrAction(self, old, new,
                                      descr="output / change output filetype"))

    def autosave_cb(self, widget, tf):
        old = dict(auto_save=self._auto_save)
        self._auto_save = tf
        new = dict(auto_save=self._auto_save)
        self.pipeline.push(AttrAction(self, old, new,
                                      descr="output / change auto save"))

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)

        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        output_folder = self._output_folder
        if output_folder is None:
            output_folder = self.pipeline.get('input_folder', '.')
        if self.gui_up:
            self.w.output_folder.set_text(output_folder)

        filename = self.pipeline.get('input_filename', '')
        if len(filename) > 0:
            f_name, f_ext = os.path.splitext(filename)
            # TODO: paramaterize the mangled filename
            output_filename = f_name + '-pipe'
            if self.gui_up:
                self.w.output_filename.set_text(output_filename)

        if self._auto_save:
            self.save_as_cb(self.w.save)

        self.pipeline.send(res_np=data)

    def save_as_cb(self, widget):
        data = self.pipeline.get_data(self)

        dct = self.convert_info[self._output_filetype]
        path = os.path.join(self.w.output_folder.get_text().strip(),
                            self.w.output_filename.get_text().strip() +
                            dct['ext'])
        try:
            self._save_as(path, data, format=dct['format'])

        except Exception as e:
            self.logger.error("Error saving file '{}': {}".format(path, e),
                              exc_info=True)

    def _save_as(self, path, data, format='jpeg', quality=90):
        # get the color profile passed through the pipeline
        output_profile = self.pipeline.get('icc_output_profile', None)
        if output_profile is not None:
            profile_file = os.path.join(ginga_home, "profiles",
                                        output_profile + ".icc")
            with open(profile_file, 'rb') as in_f:
                profile = in_f.read()
        else:
            profile = None

        if format == 'jpeg':
            img = Image.fromarray(data[:, :, 0:3].astype(np.uint8))
            img.save(path, format='jpeg', quality=quality,
                     icc_profile=profile)

        elif format == 'png':
            # alpha layer OK for PNG
            img = Image.fromarray(data.astype(np.uint8))
            img.save(path, format='png', quality=quality,
                     icc_profile=profile)

        elif format == 'fits':
            if data.dtype == np.dtype(np.uint8):
                # only if data is RGB--seems to be necessary for FITS
                data = np.moveaxis(data, 2, 0)
            hdu = fits.PrimaryHDU(data)
            # TODO: carry over metadata + some new metadata
            hdu_l = [hdu]
            if profile is not None:
                # save ICC profile as a secondary HDU
                hdu.header.set('HASPROF', True,
                               "Embedded ICC profile in other HDU")
                prof_hdu = fits.ImageHDU(np.frombuffer(profile,
                                                       dtype=np.uint8))
                prof_hdu.header.set('ICCPROF', output_profile,
                                    "Embedded ICC profile in this HDU")
                prof_hdu.name = 'ICCPROF'
                hdu_l.append(prof_hdu)
            img = fits.HDUList(hdu_l)
            img.writeto(path)

    def _get_state(self):
        return dict(output_folder=self._output_folder,
                    output_filetype=self._output_filetype,
                    auto_save=self._auto_save)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.output_folder = d['output_folder']
        self.output_filetype = d['output_filetype']
        self.auto_save = d['auto_save']



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``IRAF`` plugin implements a remote control interface for the Ginga FITS
viewer from an IRAF session.  In particular, it supports the use of the
IRAF ``display`` and ``imexamine`` commands.

**Plugin Type: Global**

``IRAF`` is a global plugin.  Only one instance can be opened.

**Usage**

The ``IRAF`` plugin allows Ginga to interoperate with IRAF in a manner
similar to IRAF and DS9.  The following IRAF commands are supported:
``imexamine``, ``rimcursor``, ``display``, and ``tvmark``.

To use the ``IRAF`` plugin, first make sure the environment variable ``IMTDEV``
is set appropriately, e.g.::

        $ export IMTDEV=inet:45005

or::

        $ export IMTDEV=unix:/tmp/.imtg45

If the environment variable is not set, Ginga will default to that used
by IRAF.

Use the "Start IRAF" menu item from the "Plugins" menu.
The GUI for the ``IRAF`` plugin will appear in the tabs on the right.

It can be more convenient to load images via Ginga than IRAF.  From
Ginga, you can load images via drag and drop or via the ``FBrowser``
plugin and then use ``imexamine`` from IRAF to do analysis tasks on
them.  You can also use the ``display`` command from IRAF to show
images already loaded in IRAF in Ginga, and then use ``imexamine`` to
select areas graphically for analysis.

When using ``imexamine`` or ``rimcursor``, the plugin disables
normal UI processing on the channel image so that keystrokes,
etc. normally caught by Ginga are passed through to IRAF.  You can
toggle back and forth between local Ginga control (e.g., keystrokes to
zoom and pan the image, or apply cut levels, etc.) and IRAF control
using the radio buttons at the top of the tab.

IRAF deals with images in enumerated "frames", whereas Ginga uses
named channels.  The bottom of the ``IRAF`` plugin GUI will show the mapping
from Ginga channels to IRAF frames.

NOTE: this plugin requires the IIS_DataListener module, which should
be in the same place that you found this plugin.

"""
import os
import threading
import time
import socket
import array
import queue as Queue

import numpy as np

from ginga import GingaPlugin, AstroImage
from ginga import cmap, imap
from ginga.gw import Widgets
from ginga.misc import Bunch

# XImage protocol support
try:
    import IIS_DataListener as iis
except ImportError:  # Need this for Sphinx build
    pass

__all__ = ['IRAF']


class IRAF(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(IRAF, self).__init__(fv)

        self.keyqueue = Queue.Queue()
        self.keyevent = threading.Event()
        self.keymap = {
            'comma': ',',
        }
        self.ctrldown = False

        self.layertag = 'iraf-canvas'
        # this will be set in initialize()
        self.canvas = None
        self.dc = fv.get_draw_classes()

        self.addr = iis.get_interface()

        self.ev_quit = self.fv.ev_quit
        self.data_task = None

        # Holds frame buffers
        self.fb = {}
        self.current_frame = 0

        # cursor position
        self.cursor_x = 1.0
        self.cursor_y = 1.0

        self.mode = 'ginga'
        self.imexam_active = False
        self.imexam_chname = None

        # init the first frame(frame 0)
        self.init_frame(0)

        # colormap for use with IRAF displays
        self.cm_iis = cmap.ColorMap('iis_iraf', cmap_iis_iraf)
        self.im_iis = imap.get_imap('ultrasmooth')

        fv.add_callback('add-channel', self.add_channel)
        fv.add_callback('delete-channel', self.delete_channel)
        #fv.set_callback('channel-change', self.focus_cb)

        self.gui_up = False

    def build_gui(self, container):

        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(False)
        ## canvas.set_callback('none-move', self.cursormotion)
        canvas.add_callback('key-press', self.window_key_press)
        canvas.add_callback('key-release', self.window_key_release)
        self.canvas = canvas

        vbox = Widgets.VBox()

        fr = Widgets.Frame("IRAF")

        captions = [
            ("Addr:", 'label', "Addr", 'llabel', 'Restart', 'button'),
            ("Set Addr:", 'label', "Set Addr", 'entry'),
            ("Control", 'hbox'),
            ("Channel:", 'label', 'Channel', 'llabel'),
        ]
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        addr = str(self.addr.name)
        b.addr.set_text(addr)
        b.restart.set_tooltip("Restart the server")
        b.restart.add_callback('activated', self.restart_cb)

        b.set_addr.set_length(100)
        b.addr.set_text(addr)
        b.set_addr.set_tooltip("Set address to run remote control server")
        b.set_addr.add_callback('activated', self.set_addr_cb)

        self.w.mode_d = {}
        btn1 = Widgets.RadioButton("Ginga")
        btn1.set_state(True)
        btn1.add_callback('activated', lambda w, val: self.switch_mode('ginga'))
        self.w.mode_d['ginga'] = btn1
        self.w.control.add_widget(btn1)
        btn2 = Widgets.RadioButton("IRAF", group=btn1)
        btn2.add_callback('activated', lambda w, val: self.switch_mode('iraf'))
        self.w.mode_d['iraf'] = btn2
        self.w.control.add_widget(btn2)

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        fr = Widgets.Frame("Frame/Channel")

        lbl = Widgets.Label("")
        self.w.frch = lbl

        fr.set_widget(lbl)
        vbox.add_widget(fr, stretch=0)

        # stretch
        vbox.add_widget(Widgets.Label(''), stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(4)
        btns.set_border_width(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(btns)

        container.add_widget(vbox, stretch=1)
        self.gui_up = True

        fmap = self.get_channel_frame_mapping()
        self.update_chinfo(fmap)

    def update_chinfo(self, fmap):
        if not self.gui_up:
            return
        # Update the GUI with the new frame/channel mapping
        fmap.sort(key=lambda x: x[1])

        s = ["%2d: %s" % (num, name) for (name, num) in fmap]
        self.w.frch.set_text("\n".join(s))

    def _set_mode(self, mode_str, chname):
        mode_str = mode_str.lower()
        self.w.mode_d[mode_str].set_state(True)
        self.w.channel.set_text(chname)

        self.switch_mode(mode_str)

    def set_mode(self, mode_str, chname):
        self.imexam_chname = chname
        self.fv.gui_do(self._set_mode, mode_str, chname)

    def toggle_mode(self):
        isIRAF = self.w.mode_d['iraf'].get_state()
        chname = self.imexam_chname
        if isIRAF:
            self.logger.info("setting mode to Ginga")
            self.set_mode('Ginga', chname)
        else:
            self.logger.info("setting mode to IRAF")
            self.set_mode('IRAF', chname)

    def add_channel(self, viewer, chinfo):
        self.logger.debug("channel %s added." % (chinfo.name))

        n = self.channel_to_frame(chinfo.name)
        if n is None:
            found = len(self.fb)
            for n, fb in self.fb.items():
                if fb.chname is None:
                    found = n
            fb = self.init_frame(found)
            fb.chname = chinfo.name

        fmap = self.get_channel_frame_mapping()
        self.fv.gui_do(self.update_chinfo, fmap)

    def delete_channel(self, viewer, chinfo):
        self.logger.debug("delete channel %s" % (chinfo.name))

        n = self.channel_to_frame(chinfo.name)
        if n is not None:
            self.fb[n].chname = None

        fmap = self.get_channel_frame_mapping()
        self.fv.gui_do(self.update_chinfo, fmap)

    def switch_mode(self, mode_str):
        mode_str = mode_str.lower()
        chname = self.imexam_chname
        chinfo = self.fv.get_channel(chname)

        if mode_str == 'iraf':
            self.ui_disable(chinfo.fitsimage)
        else:
            self.ui_enable(chinfo.fitsimage)

    def start(self):
        try:
            if self.addr.prot == 'unix':
                os.remove(self.addr.path)
        except Exception:
            pass

        # start the data listener task, if appropriate
        ev_quit = threading.Event()
        self.data_task = iis.IIS_DataListener(
            self.addr, controller=self,
            ev_quit=ev_quit, logger=self.logger)
        self.fv.nongui_do(self.data_task.mainloop)

    def stop(self):
        if self.data_task:
            self.data_task.stop()
        self.gui_up = False

    def restart_cb(self, w):
        # restart server
        if self.data_task:
            self.data_task.stop()
        self.start()

    def set_addr_cb(self, w):
        # get and parse address
        addr = w.get_text()
        self.addr = iis.get_interface(addr=addr)
        addr = str(self.addr.name)
        self.w.addr.set_text(addr)

    def channel_to_frame(self, chname):
        for n, fb in self.fb.items():
            if fb.chname == chname:
                return n
        return None

    def get_channel_frame_mapping(self):
        l = [(fb.chname, n + 1) for n, fb in self.fb.items()]
        return l

    def redo(self, channel, image):
        """Called when a new image appears in the channel."""
        if not self.gui_up:
            return
        # check if this is an image we received from IRAF or
        # one that was loaded locally.
        ct = image.get('ct', None)
        if ct is not None:
            # This image was sent by IRAF--we don't need to
            # construct extra fb information for it
            return

        n = self.channel_to_frame(channel.name)
        if n is None:
            return
        self.logger.debug("new image, frame is %d" % (n))
        fb = self.get_frame(n)
        newpath = image.get('path', 'NO_PATH')
        host = socket.getfqdn()
        newhost = image.get('host', host)
        # protocol has a bizarre 16-char limit on hostname
        newhost = newhost[:16]

        # this is just a placeholder so that IIS_RequestHandler will
        # report something in this buffer
        fb.buffer = array.array('B', ' ')

        # Update IRAF "wcs" info so that IRAF can load this image

        #print "filling wcs info"
        fb.ct = iis.coord_tran()
        image.set(ct=ct)

        # iis version 1 data
        fb.ct.valid = 1
        fb.ct.a = 1
        fb.ct.b = 0
        fb.ct.c = 0
        fb.ct.d = 1
        fb.ct.tx = 0
        fb.ct.ty = 0
        fb.ct.z1 = 0
        fb.ct.z2 = 1
        fb.ct.zt = iis.W_UNITARY
        fb.ct.format = ''
        fb.ct.imtitle = ''

        # iis version 1+ data
        fb.ct.region = 'image'
        #x1, y1, x2, y2 = fitsimage.get_datarect()
        #wd, ht = x2-x1, y2-y1
        #x1, y1, x2, y2 = x1+1, y1+1, x2+1, y2+1
        #fb.ct.sx, fb.ct.sy = float(x1), float(y1)
        wd, ht = image.get_size()
        fb.ct.sx, fb.ct.sy = float(1), float(1)
        fb.ct.snx, fb.ct.sny = wd, ht
        fb.ct.dx, fb.ct.dy = 1, 1
        fb.ct.dnx, fb.ct.dny = wd, ht

        # ref
        newref = "!".join([newhost, newpath])
        fb.ct.ref = newref

        # TODO: we shouldn't have to know about this here...
        if (fb and fb.ct.a is not None):
            wcs = "%s\n%f %f %f %f %f %f %f %f %d\n" % (
                fb.ct.imtitle, fb.ct.a, fb.ct.b, fb.ct.c, fb.ct.d,
                fb.ct.tx, fb.ct.ty, fb.ct.z1, fb.ct.z2, fb.ct.zt)
        else:
            wcs = "[NOSUCHWCS]\n"
        if (fb and fb.ct.sx is not None):
            mapping = "%s %f %f %d %d %d %d %d %d\n%s\n" % (
                fb.ct.region, fb.ct.sx, fb.ct.sy, fb.ct.snx, fb.ct.sny,
                fb.ct.dx, fb.ct.dy, fb.ct.dnx, fb.ct.dny, fb.ct.ref)
        else:
            mapping = ""
        fb.wcs = wcs + mapping
        self.logger.debug("filled wcs info")

    # ------ BEGIN (methods called by IIS server) ----------------
    def init_frame(self, n):
        """
        NOTE: this is called from the IIS_RequestHandler
        """
        self.logger.debug("initializing frame %d" % (n))
        # create the frame, if needed
        try:
            fb = self.get_frame(n)
        except KeyError:
            fb = iis.framebuffer()
            self.fb[n] = fb

        fb.width = None
        fb.height = None
        fb.wcs = ''
        fb.image = None
        fb.bitmap = None
        fb.zoom = 1.0
        fb.buffer = array.array('B')
        fb.ct = iis.coord_tran()
        #fb.chname = None
        return fb

    def get_frame(self, n):
        """
        NOTE: this is called from the IISRequestHandler
        Will raise KeyError if frame is not initialized.
        """
        return self.fb[n]

    def set_frame(self, n):
        """
        NOTE: this is called from the IISRequestHandler
        """
        self.current_frame = n

    def display(self, frame, width, height, reverse=False):
        """
        NOTE: this is called from the IISRequestHandler
        """

        fb = self.get_frame(frame)
        self.current_frame = frame

        if reverse:
            fb.buffer.reverse()

        # frames are indexed from 1 in IRAF
        chname = fb.chname
        if chname is None:
            chname = 'Frame%d' % (frame + 1)
            fb.chname = chname

        self.logger.debug("display to %s" % (chname))

        try:
            data = fb.buffer
            byteswap = False  # noqa
            dims = (fb.height, fb.width)
            dtype = np.uint8
            metadata = {}

            image = IRAF_AstroImage(logger=self.logger)
            #image.load_buffer(fb.buffer, dims, dtype, byteswap=byteswap,
            #                  metadata=metadata)
            data = np.fromstring(fb.buffer, dtype=dtype)
            data = data.reshape(dims)
            # Image comes in from IRAF flipped for screen display
            data = np.flipud(data)
            image.set_data(data, metadata=metadata)
            # Save coordinate transform info
            image.set(ct=fb.ct)

            # make up a name (is there a protocol slot for the name?)
            fitsname = str(time.time())

            # extract path from ref
            oldref = fb.ct.ref
            items = oldref.split('!')
            host = items[0]
            path = '!'.join(items[1:])

            image.set(name=fitsname, path=path, host=host)
            #image.update_keywords(header)

        except Exception as e:
            # Some kind of error unpacking the data
            errmsg = "Error creating image data for '%s': %s" % (
                chname, str(e))
            self.logger.error(errmsg)
            raise GingaPlugin.PluginError(errmsg)

        # Do the GUI bits as the GUI thread
        self.fv.gui_do(self._gui_display_image, fitsname, image, chname)

    def _gui_display_image(self, fitsname, image, chname):

        if not self.fv.has_channel(chname):
            chinfo = self.fv.add_channel(chname)
        else:
            chinfo = self.fv.get_channel(chname)
        fitsimage = chinfo.fitsimage

        # Set the RGB mapping appropriate for IIS/IRAF
        rgbmap = fitsimage.get_rgbmap()
        rgbmap.set_imap(self.im_iis, callback=False)
        rgbmap.set_cmap(self.cm_iis, callback=False)
        rgbmap.set_hash_size(65535, callback=False)
        rgbmap.set_hash_algorithm('linear', callback=True)

        # various settings that should apply
        settings = fitsimage.get_settings()
        settings.set_dict(dict(autocuts='off', flip_x=False, flip_y=False,
                               swap_xy=False, rot_deg=0.0), callback=True)

        # Set cut levels
        fitsimage.cut_levels(0.0, 255.0, no_reset=True)

        # Enqueue image to display datasrc
        self.fv.add_image(fitsname, image, chname=chname)
        self.fv.ds.raise_tab('IRAF')

    def get_cursor(self):
        self.logger.info("get_cursor() called")
        chinfo = self.fv.get_channel_info()
        # Find out which frame we are looking at
        #frame = self.current_frame
        frame = self.channel_to_frame(chinfo.name)
        fitsimage = chinfo.fitsimage
        last_x, last_y = fitsimage.get_last_data_xy()

        # Correct for surrounding framebuffer
        image = fitsimage.get_image()
        if isinstance(image, IRAF_AstroImage):
            last_x, last_y = image.get_corrected_xy(last_x, last_y)

        res = Bunch.Bunch(x=last_x, y=last_y, frame=frame)
        return res

    def get_keystroke(self):
        self.logger.info("get_keystroke() called")
        chinfo = self.fv.get_channel_info()
        fitsimage = chinfo.fitsimage
        # Find out which frame we are looking at
        #frame = self.current_frame
        frame = self.channel_to_frame(chinfo.name)
        image = fitsimage.get_image()  # noqa

        self.start_imexamine(fitsimage, chinfo.name)

        self.keyevent.wait()
        evt = self.keyqueue.get()

        self.stop_imexamine(fitsimage, chinfo.name)

        res = Bunch.Bunch(x=evt.x, y=evt.y, key=evt.key, frame=frame)
        return res

    def set_cursor(self, x, y):
        self.logger.info("TODO: set_cursor() called")

    # ------ END (methods called by IIS server) ----------------

    def ui_disable(self, fitsimage):
        # NOTE: can't disable main canvas ui because it won't propagate
        # events to layered canvases
        #fitsimage.ui_set_active(False)
        #self.canvas.ui_set_active(True)
        self.mode = 'iraf'

    def ui_enable(self, fitsimage):
        fitsimage.ui_set_active(True)
        #self.canvas.ui_set_active(False)
        self.mode = 'ginga'

    def start_imexamine(self, fitsimage, chname):
        self.logger.info("STARTING")
        # Turn off regular UI processing in the frame
        self.canvas.set_surface(fitsimage)
        # insert layer if it is not already
        try:
            fitsimage.get_object_by_tag(self.layertag)

        except KeyError:
            # Add canvas layer
            fitsimage.add(self.canvas, tag=self.layertag)
        self.canvas.ui_set_active(True)

        self.imexam_active = True
        self.set_mode('IRAF', chname)
        self.fv.gui_do(self.fv.ds.raise_tab, 'IRAF')
        self.logger.info("FINISHING")

    def stop_imexamine(self, fitsimage, chname):
        self.logger.info("STARTING")
        self.imexam_active = False
        self.set_mode('Ginga', chname)
        self.logger.info("FINISHING")

    def window_key_press(self, canvas, keyname):
        if not self.imexam_active:
            return False
        self.logger.info("key pressed: %s" % (keyname))
        if len(keyname) > 1:
            if keyname in ('shift_l', 'shift_r'):
                # ignore these keystrokes
                return False
            elif keyname in ('control_l', 'control_r'):
                # control key combination
                self.ctrldown = True
                return False
            elif keyname == 'space':
                self.toggle_mode()
                return True
            keyname = self.keymap.get(keyname, '?')

        if self.mode != 'iraf':
            return False

        if self.ctrldown:
            if keyname == 'd':
                # User typed ^D
                keyname = chr(4)

        # Get cursor position
        # TODO: not sure that this attribute will be permanent...
        fitsimage = canvas.viewer
        last_x, last_y = fitsimage.get_last_data_xy()

        # Correct for surrounding framebuffer
        image = fitsimage.get_image()
        if isinstance(image, IRAF_AstroImage):
            last_x, last_y = image.get_corrected_xy(last_x, last_y)

        # Get frame info
        #frame = self.current_frame
        chname = self.fv.get_channel_name(fitsimage)
        chinfo = self.fv.get_channel(chname)
        frame = self.channel_to_frame(chinfo.name)

        # add framebuffer information if it is not there already
        self.redo(chinfo, image)

        self.keyqueue.put(Bunch.Bunch(x=last_x, y=last_y, key=keyname,
                                      frame=frame))
        self.keyevent.set()
        return True

    def window_key_release(self, canvas, keyname):
        if not self.imexam_active:
            return False
        self.logger.info("key released: %s" % (keyname))
        if len(keyname) > 1:
            if keyname in ('control_l', 'control_r'):
                # control key combination
                self.ctrldown = False

        return False

    def cursormotion(self, canvas, event, data_x, data_y):
        if self.mode != 'iraf':
            return False

        chviewer = self.fv.getfocus_viewer()

        if event.state == 'move':
            self.fv.showxy(chviewer, data_x, data_y)
            return True

        return False

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'iraf'


class IRAF_AstroImage(AstroImage.AstroImage):

    def info_xy(self, data_x, data_y, settings):
        ct = self.get('ct', None)
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            x, y = int(data_x + 0.5), int(data_y + 0.5)
            value = self.get_data_xy(x, y)

            # Mapping from bytescaled values back to original values
            value = iis.wcs_pix_transform(ct, value)
        except Exception as e:
            self.logger.error("Exception getting value at %d,%d: %s" % (
                x, y, str(e)))
            value = None

        # Calculate WCS RA, if available
        try:
            # Subtract offsets of data in framebuffer and add offsets of
            # rect beginning in source
            data_x = data_x - (ct.dx - 1) + (ct.sx - 1)
            data_y = data_y - (ct.dy - 1) + (ct.sy - 1)

            #ra_deg, dec_deg = wcs_coord_transform(ct, data_x, data_y)

            #ra_txt, dec_txt = self.wcs.deg2fmt(ra_deg, dec_deg, 'str')
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        # Note: FITS coordinates are 1-based, whereas numpy FITS arrays
        # are 0-based
        ra_lbl, dec_lbl = chr(945), chr(948)
        fits_x, fits_y = data_x + 1, data_y + 1

        info = Bunch.Bunch(itype='astro', data_x=data_x, data_y=data_y,
                           fits_x=fits_x, fits_y=fits_y,
                           x=fits_x, y=fits_y,
                           ra_txt=ra_txt, dec_txt=dec_txt,
                           ra_lbl=ra_lbl, dec_lbl=dec_lbl,
                           value=value)
        return info

    def get_corrected_xy(self, data_x, data_y):
        ct = self.get('ct', None)
        if ct is not None:
            # Subtract offsets of data in framebuffer and add offsets of
            # rect beginning in source
            data_x = data_x - (ct.dx - 1) + (ct.sx - 1)
            data_y = data_y - (ct.dy - 1) + (ct.sy - 1)
        return data_x, data_y


# this is a specialized color map for use with IRAF displays
#
cmap_iis_iraf = (
    (0.000000, 0.000000, 0.000000),   # 0: black
    (0.004975, 0.004975, 0.004975),   # 1-200: frame buffer greyscale values
    (0.009950, 0.009950, 0.009950),
    (0.014925, 0.014925, 0.014925),
    (0.019900, 0.019900, 0.019900),
    (0.024876, 0.024876, 0.024876),
    (0.029851, 0.029851, 0.029851),
    (0.034826, 0.034826, 0.034826),
    (0.039801, 0.039801, 0.039801),
    (0.044776, 0.044776, 0.044776),
    (0.049751, 0.049751, 0.049751),
    (0.054726, 0.054726, 0.054726),
    (0.059701, 0.059701, 0.059701),
    (0.064677, 0.064677, 0.064677),
    (0.069652, 0.069652, 0.069652),
    (0.074627, 0.074627, 0.074627),
    (0.079602, 0.079602, 0.079602),
    (0.084577, 0.084577, 0.084577),
    (0.089552, 0.089552, 0.089552),
    (0.094527, 0.094527, 0.094527),
    (0.099502, 0.099502, 0.099502),
    (0.104478, 0.104478, 0.104478),
    (0.109453, 0.109453, 0.109453),
    (0.114428, 0.114428, 0.114428),
    (0.119403, 0.119403, 0.119403),
    (0.124378, 0.124378, 0.124378),
    (0.129353, 0.129353, 0.129353),
    (0.134328, 0.134328, 0.134328),
    (0.139303, 0.139303, 0.139303),
    (0.144279, 0.144279, 0.144279),
    (0.149254, 0.149254, 0.149254),
    (0.154229, 0.154229, 0.154229),
    (0.159204, 0.159204, 0.159204),
    (0.164179, 0.164179, 0.164179),
    (0.169154, 0.169154, 0.169154),
    (0.174129, 0.174129, 0.174129),
    (0.179104, 0.179104, 0.179104),
    (0.184080, 0.184080, 0.184080),
    (0.189055, 0.189055, 0.189055),
    (0.194030, 0.194030, 0.194030),
    (0.199005, 0.199005, 0.199005),
    (0.203980, 0.203980, 0.203980),
    (0.208955, 0.208955, 0.208955),
    (0.213930, 0.213930, 0.213930),
    (0.218905, 0.218905, 0.218905),
    (0.223881, 0.223881, 0.223881),
    (0.228856, 0.228856, 0.228856),
    (0.233831, 0.233831, 0.233831),
    (0.238806, 0.238806, 0.238806),
    (0.243781, 0.243781, 0.243781),
    (0.248756, 0.248756, 0.248756),
    (0.253731, 0.253731, 0.253731),
    (0.258706, 0.258706, 0.258706),
    (0.263682, 0.263682, 0.263682),
    (0.268657, 0.268657, 0.268657),
    (0.273632, 0.273632, 0.273632),
    (0.278607, 0.278607, 0.278607),
    (0.283582, 0.283582, 0.283582),
    (0.288557, 0.288557, 0.288557),
    (0.293532, 0.293532, 0.293532),
    (0.298507, 0.298507, 0.298507),
    (0.303483, 0.303483, 0.303483),
    (0.308458, 0.308458, 0.308458),
    (0.313433, 0.313433, 0.313433),
    (0.318408, 0.318408, 0.318408),
    (0.323383, 0.323383, 0.323383),
    (0.328358, 0.328358, 0.328358),
    (0.333333, 0.333333, 0.333333),
    (0.338308, 0.338308, 0.338308),
    (0.343284, 0.343284, 0.343284),
    (0.348259, 0.348259, 0.348259),
    (0.353234, 0.353234, 0.353234),
    (0.358209, 0.358209, 0.358209),
    (0.363184, 0.363184, 0.363184),
    (0.368159, 0.368159, 0.368159),
    (0.373134, 0.373134, 0.373134),
    (0.378109, 0.378109, 0.378109),
    (0.383085, 0.383085, 0.383085),
    (0.388060, 0.388060, 0.388060),
    (0.393035, 0.393035, 0.393035),
    (0.398010, 0.398010, 0.398010),
    (0.402985, 0.402985, 0.402985),
    (0.407960, 0.407960, 0.407960),
    (0.412935, 0.412935, 0.412935),
    (0.417910, 0.417910, 0.417910),
    (0.422886, 0.422886, 0.422886),
    (0.427861, 0.427861, 0.427861),
    (0.432836, 0.432836, 0.432836),
    (0.437811, 0.437811, 0.437811),
    (0.442786, 0.442786, 0.442786),
    (0.447761, 0.447761, 0.447761),
    (0.452736, 0.452736, 0.452736),
    (0.457711, 0.457711, 0.457711),
    (0.462687, 0.462687, 0.462687),
    (0.467662, 0.467662, 0.467662),
    (0.472637, 0.472637, 0.472637),
    (0.477612, 0.477612, 0.477612),
    (0.482587, 0.482587, 0.482587),
    (0.487562, 0.487562, 0.487562),
    (0.492537, 0.492537, 0.492537),
    (0.497512, 0.497512, 0.497512),
    (0.502488, 0.502488, 0.502488),
    (0.507463, 0.507463, 0.507463),
    (0.512438, 0.512438, 0.512438),
    (0.517413, 0.517413, 0.517413),
    (0.522388, 0.522388, 0.522388),
    (0.527363, 0.527363, 0.527363),
    (0.532338, 0.532338, 0.532338),
    (0.537313, 0.537313, 0.537313),
    (0.542289, 0.542289, 0.542289),
    (0.547264, 0.547264, 0.547264),
    (0.552239, 0.552239, 0.552239),
    (0.557214, 0.557214, 0.557214),
    (0.562189, 0.562189, 0.562189),
    (0.567164, 0.567164, 0.567164),
    (0.572139, 0.572139, 0.572139),
    (0.577114, 0.577114, 0.577114),
    (0.582090, 0.582090, 0.582090),
    (0.587065, 0.587065, 0.587065),
    (0.592040, 0.592040, 0.592040),
    (0.597015, 0.597015, 0.597015),
    (0.601990, 0.601990, 0.601990),
    (0.606965, 0.606965, 0.606965),
    (0.611940, 0.611940, 0.611940),
    (0.616915, 0.616915, 0.616915),
    (0.621891, 0.621891, 0.621891),
    (0.626866, 0.626866, 0.626866),
    (0.631841, 0.631841, 0.631841),
    (0.636816, 0.636816, 0.636816),
    (0.641791, 0.641791, 0.641791),
    (0.646766, 0.646766, 0.646766),
    (0.651741, 0.651741, 0.651741),
    (0.656716, 0.656716, 0.656716),
    (0.661692, 0.661692, 0.661692),
    (0.666667, 0.666667, 0.666667),
    (0.671642, 0.671642, 0.671642),
    (0.676617, 0.676617, 0.676617),
    (0.681592, 0.681592, 0.681592),
    (0.686567, 0.686567, 0.686567),
    (0.691542, 0.691542, 0.691542),
    (0.696517, 0.696517, 0.696517),
    (0.701493, 0.701493, 0.701493),
    (0.706468, 0.706468, 0.706468),
    (0.711443, 0.711443, 0.711443),
    (0.716418, 0.716418, 0.716418),
    (0.721393, 0.721393, 0.721393),
    (0.726368, 0.726368, 0.726368),
    (0.731343, 0.731343, 0.731343),
    (0.736318, 0.736318, 0.736318),
    (0.741294, 0.741294, 0.741294),
    (0.746269, 0.746269, 0.746269),
    (0.751244, 0.751244, 0.751244),
    (0.756219, 0.756219, 0.756219),
    (0.761194, 0.761194, 0.761194),
    (0.766169, 0.766169, 0.766169),
    (0.771144, 0.771144, 0.771144),
    (0.776119, 0.776119, 0.776119),
    (0.781095, 0.781095, 0.781095),
    (0.786070, 0.786070, 0.786070),
    (0.791045, 0.791045, 0.791045),
    (0.796020, 0.796020, 0.796020),
    (0.800995, 0.800995, 0.800995),
    (0.805970, 0.805970, 0.805970),
    (0.810945, 0.810945, 0.810945),
    (0.815920, 0.815920, 0.815920),
    (0.820896, 0.820896, 0.820896),
    (0.825871, 0.825871, 0.825871),
    (0.830846, 0.830846, 0.830846),
    (0.835821, 0.835821, 0.835821),
    (0.840796, 0.840796, 0.840796),
    (0.845771, 0.845771, 0.845771),
    (0.850746, 0.850746, 0.850746),
    (0.855721, 0.855721, 0.855721),
    (0.860697, 0.860697, 0.860697),
    (0.865672, 0.865672, 0.865672),
    (0.870647, 0.870647, 0.870647),
    (0.875622, 0.875622, 0.875622),
    (0.880597, 0.880597, 0.880597),
    (0.885572, 0.885572, 0.885572),
    (0.890547, 0.890547, 0.890547),
    (0.895522, 0.895522, 0.895522),
    (0.900498, 0.900498, 0.900498),
    (0.905473, 0.905473, 0.905473),
    (0.910448, 0.910448, 0.910448),
    (0.915423, 0.915423, 0.915423),
    (0.920398, 0.920398, 0.920398),
    (0.925373, 0.925373, 0.925373),
    (0.930348, 0.930348, 0.930348),
    (0.935323, 0.935323, 0.935323),
    (0.940299, 0.940299, 0.940299),
    (0.945274, 0.945274, 0.945274),
    (0.950249, 0.950249, 0.950249),
    (0.955224, 0.955224, 0.955224),
    (0.960199, 0.960199, 0.960199),
    (0.965174, 0.965174, 0.965174),
    (0.970149, 0.970149, 0.970149),
    (0.975124, 0.975124, 0.975124),
    (0.980100, 0.980100, 0.980100),
    (0.985075, 0.985075, 0.985075),
    (0.990050, 0.990050, 0.990050),
    (0.995025, 0.995025, 0.995025),   # 200: end of IRAF greyscale
    (1.000000, 1.000000, 1.000000),   # 201: white
    (0.000000, 0.000000, 0.000000),   # 202: black
    (1.000000, 1.000000, 1.000000),   # 203: white
    (1.000000, 0.000000, 0.000000),   # 204: red
    (0.000000, 1.000000, 0.000000),   # 205: green
    (0.000000, 0.000000, 1.000000),   # 206: blue
    (1.000000, 1.000000, 0.000000),   # 207: yellow
    (0.000000, 1.000000, 1.000000),   # 208: cyan
    (1.000000, 0.000000, 1.000000),   # 209: magenta
    (1.000000, 0.498039, 0.313725),   # 210: coral
    (0.690196, 0.188235, 0.376470),   # 211: maroon
    (1.000000, 0.647058, 0.000000),   # 212: orange
    (0.941176, 0.901960, 0.549019),   # 213: khaki
    (0.854901, 0.439215, 0.839215),   # 214: orchid
    (0.250980, 0.878431, 0.815686),   # 215: turquoise
    (0.933333, 0.509803, 0.933333),   # 216: violet
    (0.960784, 0.870588, 0.701960),   # 217: wheat
    (1.000000, 0.000000, 0.000000),   # 218-254: reserved
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (1.000000, 0.000000, 0.000000),
    (0.000000, 0.000000, 0.000000),   # 255: black
)


# END
